/* File *******************************************************************
*					North Carolina State University
*					Electronics Research Laboratory
*
* FILENAME :	analyze.c
*
* DESCRIPTION :	This file provides the basic structure to analyze a s2ibis2
*				component data structure and generate the appropriate data
*				for the models within that structure.  It acts on each pin,
*				in the component, determining whether Spice modeling is 
*				needed for the pin, and which types of simulations are
*				necessary.  It then calls the appropriate routines, which
*				are found in the file spicegen.c.
*
*
* FUNCTIONS :	findSupplyPins
*				analyzePin
*				analyzeComponent
*
*
* NOTES :
*
*H*/

/*==== Included Files ===================================================*/
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "s2ibis.h"
#include "s2ilink.h"


/*==== External References ==============================================*/

/*---- Function Prototypes ----------------------------------------------*/
ibisPin_Pt getMatchingPin( char[], ibisPin_Pt );

int generateVICurve( ibisVItable_Pt *, ibisPin_Pt, ibisPin_Pt, ibisPin_Pt, 
					 ibisPin_Pt, ibisPin_Pt, ibisTypMinMax_t, ibisTypMinMax_t,					 ibisTypMinMax_t, double, double, int, int, char *, 
					 char *, int, int, int, int );

/*==== Global Declarations ==============================================*/

/*---- Macros -----------------------------------------------------------*/
#define THIS_PIN_NEEDS_ANALYSIS( modelName ) \
		     strcmp( modelName, "POWER" ) && \
			 strcmp( modelName, "GND" ) && \
			 strcmp( modelName, "NC" ) 

#define THIS_MODEL_NEEDS_PULLUP_DATA( modelType ) \
			( modelType == MODEL_TYPE_OUTPUT ) || \
			( modelType == MODEL_TYPE_3_STATE ) || \
			( modelType == MODEL_TYPE_I_O ) || \
			( modelType == MODEL_TYPE_OPEN_SOURCE ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_SOURCE ) || \
			( modelType == MODEL_TYPE_OUTPUT_ECL ) || \
			( modelType == MODEL_TYPE_I_O_ECL ) 
			
#define THIS_MODEL_NEEDS_PULLDOWN_DATA( modelType ) \
			( modelType == MODEL_TYPE_OUTPUT ) || \
			( modelType == MODEL_TYPE_3_STATE ) || \
			( modelType == MODEL_TYPE_I_O ) || \
			( modelType == MODEL_TYPE_OPEN_SINK ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_SINK ) || \
			( modelType == MODEL_TYPE_OPEN_DRAIN ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_DRAIN ) || \
			( modelType == MODEL_TYPE_OUTPUT_ECL ) || \
			( modelType == MODEL_TYPE_I_O_ECL )

#define THIS_MODEL_NEEDS_POWER_CLAMP_DATA( modelType ) \
			( modelType == MODEL_TYPE_INPUT ) || \
			( modelType == MODEL_TYPE_3_STATE ) || \
			( modelType == MODEL_TYPE_I_O ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_SOURCE ) || \
			( modelType == MODEL_TYPE_INPUT_ECL ) || \
			( modelType == MODEL_TYPE_I_O_ECL ) || \
			( modelType == MODEL_TYPE_TERMINATOR ) 

#define THIS_MODEL_NEEDS_GND_CLAMP_DATA( modelType ) \
			( modelType == MODEL_TYPE_INPUT ) || \
			( modelType == MODEL_TYPE_3_STATE ) || \
			( modelType == MODEL_TYPE_I_O ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_SINK ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_DRAIN ) || \
			( modelType == MODEL_TYPE_INPUT_ECL ) || \
			( modelType == MODEL_TYPE_I_O_ECL ) || \
			( modelType == MODEL_TYPE_TERMINATOR )

#define THIS_MODEL_NEEDS_TRANSIENT_DATA( modelType ) \
			( modelType == MODEL_TYPE_OUTPUT ) || \
			( modelType == MODEL_TYPE_3_STATE ) || \
			( modelType == MODEL_TYPE_I_O ) || \
			( modelType == MODEL_TYPE_OPEN_SINK ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_SINK ) || \
			( modelType == MODEL_TYPE_OPEN_DRAIN ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_DRAIN ) || \
			( modelType == MODEL_TYPE_OPEN_SOURCE ) || \
			( modelType == MODEL_TYPE_I_O_OPEN_SOURCE ) || \
			( modelType == MODEL_TYPE_OUTPUT_ECL ) || \
			( modelType == MODEL_TYPE_I_O_ECL )


void setupVoltages(	IN	int					curveType,
					IN	ibisModel_Pt		model_P,
					OUT	ibisTypMinMax_Pt	Vcc_P,
					OUT	ibisTypMinMax_Pt	Gnd_P,
					OUT	ibisTypMinMax_Pt	sweepStart_P,
					OUT	double				*sweepRange_P,
					OUT	double				*sweepStep_P,
					OUT	double				*diodeDrop_P )

{
ibisTypMinMax_Pt	pullupRef_P,
					pulldownRef_P,
					powerClampRef_P,
					gndClampRef_P;
ibisTypMinMax_t		zeros;
double				linRange,
					minOffset,
					maxOffset;


zeros.typ = 0;
zeros.min = 0;
zeros.max = 0;

if ( IS_USE_NA( model_P->pullupRef.typ ) )
	pullupRef_P = &( model_P->voltageRange );
else
	pullupRef_P = &( model_P->pullupRef );

if ( IS_USE_NA( model_P->pulldownRef.typ ) )
	pulldownRef_P = &( zeros );
else
	pulldownRef_P = &( model_P->pulldownRef );

if ( IS_USE_NA( model_P->powerClampRef.typ ) )
	powerClampRef_P = &( model_P->voltageRange );
else
	powerClampRef_P = &( model_P->powerClampRef );

if ( IS_USE_NA( model_P->gndClampRef.typ ) )
	gndClampRef_P = &( zeros );
else
	gndClampRef_P = &( model_P->gndClampRef );


if ( IS_ECL( model_P->modelType ) )
	{
	*diodeDrop_P = 0;

	Vcc_P->typ = pullupRef_P->typ;
	Vcc_P->min = pullupRef_P->min;
	Vcc_P->max = pullupRef_P->max;

	if ( IS_USE_NA( model_P->gndClampRef.typ ) )
		{
		if ( ( Vcc_P->typ >= 4.5 ) && ( Vcc_P->typ <= 5.5 ) )
			{
			Gnd_P->typ = 0;
			Gnd_P->min = 0;
			Gnd_P->max = 0;
			}
		else
			{
			Gnd_P->typ = Vcc_P->typ - 5.2;
			Gnd_P->min = Vcc_P->min - 5.2;
			Gnd_P->max = Vcc_P->max - 5.2;
			}
		}
	else
		{
		Gnd_P->typ = gndClampRef_P->typ;
		Gnd_P->min = gndClampRef_P->min;
		Gnd_P->max = gndClampRef_P->max;
		}
	
	if ( curveType == CURVE_TYPE_POWER_CLAMP )
		{
		sweepStart_P->typ = powerClampRef_P->typ;
		sweepStart_P->min = powerClampRef_P->min;
		sweepStart_P->max = powerClampRef_P->max;

		*sweepRange_P = ECL_SWEEP_RANGE_DEFAULT;
		*sweepStep_P = SWEEP_STEP_DEFAULT;
		*sweepRange_P += *sweepStep_P / 10;
		}
	else if ( curveType == CURVE_TYPE_GND_CLAMP )
		{
		sweepStart_P->typ = gndClampRef_P->typ - ECL_SWEEP_RANGE_DEFAULT;
		sweepStart_P->min = sweepStart_P->typ;
		sweepStart_P->max = sweepStart_P->typ;

		*sweepRange_P = powerClampRef_P->typ - sweepStart_P->typ;
		if ( *sweepRange_P < 0 )
			*sweepStep_P = -SWEEP_STEP_DEFAULT;
		else
			*sweepStep_P = SWEEP_STEP_DEFAULT;

		while ( (int) ( fabs( *sweepRange_P / *sweepStep_P ) + 1 ) > 
				MAX_TABLE_SIZE )
			*sweepStep_P *= 2;

		*sweepRange_P += *sweepStep_P / 10;
		}
	else
		{
		sweepStart_P->typ = pullupRef_P->typ - ECL_SWEEP_RANGE_DEFAULT;
		sweepStart_P->min = sweepStart_P->typ;
		sweepStart_P->max = sweepStart_P->typ;

		if ( ( curveType == CURVE_TYPE_PULLUP ) || 
			 ( curveType == CURVE_TYPE_DISABLED_PULLUP ) )
			{
			maxOffset = Vcc_P->max - Vcc_P->typ;
			minOffset = Vcc_P->min - Vcc_P->typ;
			
			sweepStart_P->max = sweepStart_P->max + maxOffset;
			sweepStart_P->min = sweepStart_P->min + minOffset;
			}

		*sweepRange_P = ECL_SWEEP_RANGE_DEFAULT;
		*sweepStep_P = SWEEP_STEP_DEFAULT;
		*sweepRange_P += *sweepStep_P / 10;
		}
	}
else
	{
	if ( ( model_P->modelType == MODEL_TYPE_INPUT ) ||
		 ( model_P->modelType == MODEL_TYPE_TERMINATOR ) )
		linRange = powerClampRef_P->typ - gndClampRef_P->typ;
	else
		linRange = pullupRef_P->typ - pulldownRef_P->typ;
	linRange = MIN( linRange, LIN_RANGE_DEFAULT );

	if ( ( curveType == CURVE_TYPE_GND_CLAMP ) ||
		 ( curveType == CURVE_TYPE_POWER_CLAMP ) )
		{
		Vcc_P->typ = powerClampRef_P->typ;
		Vcc_P->min = powerClampRef_P->min;
		Vcc_P->max = powerClampRef_P->max;

		Gnd_P->typ = gndClampRef_P->typ;
		Gnd_P->min = gndClampRef_P->min;
		Gnd_P->max = gndClampRef_P->max;

		if ( curveType == CURVE_TYPE_GND_CLAMP )
			{
			sweepStart_P->typ = gndClampRef_P->typ - linRange;
			sweepStart_P->min = sweepStart_P->typ;
			sweepStart_P->max = sweepStart_P->typ;

			*sweepRange_P = powerClampRef_P->typ - sweepStart_P->typ;
			}
		else
			{
			sweepStart_P->typ = powerClampRef_P->typ;
			sweepStart_P->min = powerClampRef_P->min;
			sweepStart_P->max = powerClampRef_P->max;

			*sweepRange_P = linRange;
			}

		if ( *sweepRange_P < 0 )
			{
			*sweepStep_P = -SWEEP_STEP_DEFAULT;
			*diodeDrop_P = -DIODE_DROP_DEFAULT;
			}
		else
			{
			*sweepStep_P = SWEEP_STEP_DEFAULT;
			*diodeDrop_P = DIODE_DROP_DEFAULT;
			}

		while ( ( fabs( *sweepRange_P / *sweepStep_P ) + 1 ) > MAX_TABLE_SIZE )
			*sweepStep_P *= 2;

		*sweepRange_P += *sweepStep_P / 10;
		}
	else
		{
		Vcc_P->typ = pullupRef_P->typ;
		Vcc_P->min = pullupRef_P->min;
		Vcc_P->max = pullupRef_P->max;

		Gnd_P->typ = pulldownRef_P->typ;
		Gnd_P->min = pulldownRef_P->min;
		Gnd_P->max = pulldownRef_P->max;

		sweepStart_P->typ = pulldownRef_P->typ - linRange;
		sweepStart_P->min = sweepStart_P->typ;
		sweepStart_P->max = sweepStart_P->typ;

		if ( ( curveType == CURVE_TYPE_PULLUP ) || 
			 ( curveType == CURVE_TYPE_DISABLED_PULLUP ) )
			{
			maxOffset = Vcc_P->max - Vcc_P->typ;
			minOffset = Vcc_P->min - Vcc_P->typ;
			
			sweepStart_P->max = sweepStart_P->max + maxOffset;
			sweepStart_P->min = sweepStart_P->min + minOffset;
			}

		*sweepRange_P = pullupRef_P->typ + linRange - sweepStart_P->typ;

		if ( *sweepRange_P < 0 )
			{
			*sweepStep_P = -SWEEP_STEP_DEFAULT;
			*diodeDrop_P = -DIODE_DROP_DEFAULT;
			}
		else
			{
			*sweepStep_P = SWEEP_STEP_DEFAULT;
			*diodeDrop_P = DIODE_DROP_DEFAULT;
			}

		while ( ( fabs( ( Vcc_P->typ - Gnd_P->typ + 2 * ( *diodeDrop_P ) ) / 
				  *sweepStep_P ) + 1 ) > MAX_TABLE_SIZE )
			*sweepStep_P *= 2;

		*sweepRange_P += *sweepStep_P / 10;
		}
	}
}
				

		


/* Function ***************************************************************
*
* NAME : findSupplyPins
*
* INPUTS :	ibisPin_Pt	currentPin_P	ptr to current component
*			ibisPin_Pt	pinList_P		ptr to global pin list
*			int			hasPinMapping	flag for existence of pin map
*
*
* OUTPUTS :     0 if successful
*               1 if error 
*
*
* RETURNS : ibisPin_Pt	pullupPin_P		ptr to Vcc for pullup/down tests
*			ibisPin_Pt	pulldownPin_P	ptr to GND for pullup/down tests
*			ibisPin_Pt	powerClampPin_P ptr to Vcc for clamp tests
*			ibisPin_Pt	gndClampPin_Pt	ptr to GND for clamp tests
*
*
* NOTES / SIDE EFFECTS :	findSupplyPins searches for the correct pins
*							to attach power and ground to in order to perform
*							the pullup, pulldown, powerclamp and ground
*							clamp tests.
*F*/


int  findSupplyPins( OUT	ibisPin_Pt		*pullupPin_PP,
					 OUT	ibisPin_Pt		*pulldownPin_PP,
					 OUT	ibisPin_Pt		*powerClampPin_PP,
					 OUT	ibisPin_Pt		*gndClampPin_PP,
					 IN		ibisPin_Pt		currentPin_P,
					 IN		ibisPin_Pt		pinList_P,
					 IN		int				hasPinMapping )

{
/*==== Declarations =====================================================*/
ibisPin_Pt	temp_P = NULL;


/*==== Code =============================================================*/
if ( hasPinMapping )
	{
	/*--------------------------------------------------------------------
	*
	*	If this component has a pin map, use it to find the correct supplies.
	*   First find the power or gnd pin corresponding to the pullupRef.
	*	To do this, we search through the pin list until we find a POWER
	*	or GND pin with the same bus name in the pullupRef slot.
	*
	*-------------------------------------------------------------------*/
	if ( strcmp( currentPin_P->pullupRef, "NC" ) )
		{
		for ( temp_P = pinList_P; 
			  ( temp_P != NULL ) &&
			  ( ( strcmp( temp_P->modelName, "POWER" ) &&
				  strcmp( temp_P->modelName, "GND" ) ) ||
				strcmp( temp_P->pullupRef, currentPin_P->pullupRef ));
			  temp_P = temp_P->nextPin_P );
		if ( temp_P == NULL )
			{
			fprintf( stderr, "s2ibis2: No POWER or GND pin " );
			fprintf( stderr, "associated with bus %s.\n", 
					 currentPin_P->pullupRef );
			fprintf( stderr, "         The bus has been specified ");
			fprintf( stderr, "as a power or ground bus in the\n" );
			fprintf( stderr, "         [Pin Mapping] section of " );
			fprintf( stderr, "your input file, but it is not\n" );
			fprintf( stderr, "         connected to a pin with " );
			fprintf( stderr, "model name POWER or GND.\n" );
			return( 1 );
			}
		else
			*pullupPin_PP = temp_P;
		}
	else
		*pullupPin_PP = NULL;
	
	/*--------------------------------------------------------------------
	*
	*   Next find the power or gnd pin corresponding to the pulldownRef.
	*
	*-------------------------------------------------------------------*/
	if ( strcmp( currentPin_P->pulldownRef, "NC" ) )
		{
		for ( temp_P = pinList_P; 
			  ( temp_P != NULL ) &&
			  ( ( strcmp( temp_P->modelName, "POWER" ) &&
				  strcmp( temp_P->modelName, "GND" ) ) ||
				strcmp( temp_P->pulldownRef, currentPin_P->pulldownRef )); 
			  temp_P = temp_P->nextPin_P );
		if ( temp_P == NULL )
			{
			fprintf( stderr, "s2ibis2: No POWER or GND pin " );
			fprintf( stderr, "associated with bus %s.\n", 
					 currentPin_P->pulldownRef );
			fprintf( stderr, "         The bus has been specified ");
			fprintf( stderr, "as a power or ground bus in the\n" );
			fprintf( stderr, "         [Pin Mapping] section of " );
			fprintf( stderr, "your input file, but it is not\n" );
			fprintf( stderr, "         connected to a pin with " );
			fprintf( stderr, "model name POWER or GND.\n" );
			return( 1 );
			}
		else
			*pulldownPin_PP = temp_P;
		}
	else
		*pulldownPin_PP = NULL;
	
	/*--------------------------------------------------------------------
	*
	*   Next find the power or gnd pin corresponding to the powerClampRef.
	*
	*-------------------------------------------------------------------*/
	if ( strcmp( currentPin_P->powerClampRef, "NC" ) )
		{
		if ( strcmp( currentPin_P->powerClampRef, "" ) )
			{
			for ( temp_P = pinList_P; 
				  ( temp_P != NULL ) &&
				  ( ( strcmp( temp_P->modelName, "POWER" ) &&
					  strcmp( temp_P->modelName, "GND" ) ) ||
					strcmp( temp_P->pullupRef, 
							currentPin_P->powerClampRef ) );
				  temp_P = temp_P->nextPin_P );
			if ( temp_P == NULL )
				{
				fprintf( stderr, "s2ibis2: No POWER or GND pin " );
				fprintf( stderr, "associated with bus %s.\n", 
						 currentPin_P->powerClampRef );
				fprintf( stderr, "         The bus has been specified ");
				fprintf( stderr, "as a power or ground bus in the\n" );
				fprintf( stderr, "         [Pin Mapping] section of " );
				fprintf( stderr, "your input file, but it is not\n" );
				fprintf( stderr, "         connected to a pin with " );
				fprintf( stderr, "model name POWER or GND.\n" );
				return( 1 );
				}
			else
				*powerClampPin_PP = temp_P;
			}	
		else
			*powerClampPin_PP = *pullupPin_PP;
		}
	else
		*powerClampPin_PP = NULL;
	
	/*--------------------------------------------------------------------
	*
	*   Finally, find the power or gnd pin corresponding to the gndClampRef.
	*
	*-------------------------------------------------------------------*/
	if ( strcmp( currentPin_P->gndClampRef, "NC" ) )
		{
		if ( strcmp( currentPin_P->gndClampRef, "" ) )
			{
			for ( temp_P = pinList_P; 
				  ( temp_P != NULL ) &&
				  ( ( strcmp( temp_P->modelName, "POWER" ) &&
					  strcmp( temp_P->modelName, "GND" ) ) ||
					strcmp( temp_P->pulldownRef, currentPin_P->gndClampRef ));
				  temp_P = temp_P->nextPin_P );
			if ( temp_P == NULL )
				{
				fprintf( stderr, "s2ibis2: No POWER or GND pin " );
				fprintf( stderr, "associated with bus %s.\n", 
						 currentPin_P->gndClampRef );
				fprintf( stderr, "         The bus has been specified ");
				fprintf( stderr, "as a power or ground bus in the\n" );
				fprintf( stderr, "         [Pin Mapping] section of " );
				fprintf( stderr, "your input file, but it is not\n" );
				fprintf( stderr, "         connected to a pin with " );
				fprintf( stderr, "model name POWER or GND.\n" );
				return( 1 );
				}
			else
				*gndClampPin_PP = temp_P;
			}	
		else
			*gndClampPin_PP = *pulldownPin_PP;
		}
	else
		*gndClampPin_PP = NULL;
	}
else
	{
    /*--------------------------------------------------------------------
    *
    *   If there is no pin mapping, just use the first power and gnd pins
	*	you come to.  First find the POWER pin.
    *
    *-------------------------------------------------------------------*/
	for ( temp_P = pinList_P; 
		  ( temp_P != NULL ) &&
		  strcmp( temp_P->modelName, "POWER" );
		  temp_P = temp_P->nextPin_P );

	if ( temp_P == NULL )
		{
		fprintf( stderr,"s2ibis2: No POWER pin for component.\n" );
		fprintf( stderr,"         Please specify at least one " );
		fprintf( stderr, "pin with model name POWER.\n" );
		return( 1 );
		}
	else
		{
		*powerClampPin_PP = temp_P;
		*pullupPin_PP = temp_P;
		}
	
    /*--------------------------------------------------------------------
    *
    *   Now find the GND pin.
    *
    *-------------------------------------------------------------------*/
	for ( temp_P = pinList_P; 
		  ( temp_P != NULL ) &&
		  strcmp( temp_P->modelName, "GND" );
		  temp_P = temp_P->nextPin_P );
	if ( temp_P == NULL )
		{
		fprintf( stderr, "s2ibis2: No GND pin for component.\n" );
		fprintf( stderr,"          Please specify at least one " );
		fprintf( stderr, "pin with model name GND.\n" );
		return( 1 );
		}
	else
		{
		*gndClampPin_PP = temp_P;
		*pulldownPin_PP = temp_P;
		}
	}

return( 0 );
}


int
sortVIData(	INOUT	ibisModel_Pt	model_P, 
			IN		ibisVItable_Pt	pullupData_P, 
			IN		ibisVItable_Pt	pulldownData_P,
			IN		ibisVItable_Pt	powerClampData_P, 
			IN		ibisVItable_Pt	gndClampData_P )

{
ibisTypMinMax_t	Vcc,
				Gnd,
				sweepStart;
double			sweepRange = 0,
				sweepStep = 0,
				diodeDrop = 0;
int				numLinearPoints = 0,
				numTablePoints = 0,
				numInnerPoints = 0,
				pointsToSkip = 0,
				currentMarker = 0,
				i = 0,
				j = 0;

if ( IS_ECL( model_P->modelType ) )
	numLinearPoints = 0;
else
	numLinearPoints = NUM_PTS_LINEAR_REGION;

if ( pullupData_P != NULL )
	{
	setupVoltages( CURVE_TYPE_PULLUP, model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );

	numInnerPoints = fabs( Vcc.typ - Gnd.typ + 2*diodeDrop ) / sweepStep + 1;
	pointsToSkip = (int) ( fabs(sweepRange - (Vcc.typ - Gnd.typ + 2*diodeDrop))
				   / 2 * sweepStep * numLinearPoints );
	numTablePoints = 2 * numLinearPoints + numInnerPoints; 

	/*------------------------------------------------------------------
	*	First thing to do is to create the storage space.
	*-----------------------------------------------------------------*/
	if ( create_VItable( &( model_P->pullup ), numTablePoints ) )
		return( 1 );
	
	/*------------------------------------------------------------------
	*	Since this is a pullup curve, we need to make the table data
	*	Vcc relative, i.e. Vtable = Vcc - Vout.  We'll do this by 
	*	reading through the data and subtracting the table voltage
	*	from VccTyp.
	*-----------------------------------------------------------------*/
	for ( i = 0; i < pullupData_P->size; i++ )
		pullupData_P->VIs[i].v = Vcc.typ - pullupData_P->VIs[i].v;

	/*------------------------------------------------------------------
	*	Now we place the data into the model.  To save on the number of
	*	points stored, we only save a few of the points in the linear
	*	ranges (this only applies for non-ECL model types).  Note that 
	*	we also go through the table backward, so the points are stored 
	*	in ascending order.  First we go through the first linear range.
	*-----------------------------------------------------------------*/
	j = pullupData_P->size - 1;
	for ( i = 0; i < numLinearPoints; i ++ )
		{
		model_P->pullup->VIs[i] = pullupData_P->VIs[j];
		j -= pointsToSkip;
		}
	j += pointsToSkip;
	currentMarker = i;

	/*------------------------------------------------------------------
	*	Skip data until we get to the start of the active region.
	*-----------------------------------------------------------------*/
	while ( pullupData_P->VIs[j].v != -diodeDrop )
		j--;
	
	/*------------------------------------------------------------------
	*	Now we do the "interesting" part of the curve.
	*-----------------------------------------------------------------*/
	for ( i = currentMarker; i < currentMarker + numInnerPoints;
		  i++ )
		{
		model_P->pullup->VIs[i] = pullupData_P->VIs[j];
		j--;
		}
	currentMarker = i;

	/*------------------------------------------------------------------
	*	Now we do the second linear part of the curve.
	*-----------------------------------------------------------------*/
	j++;
	for ( i = currentMarker; i < numTablePoints; i++ )
		{
		j -= pointsToSkip;
		model_P->pullup->VIs[i] = pullupData_P->VIs[j];
		}

	/*------------------------------------------------------------------
	*	Now we apply the derating percentages.
	*-----------------------------------------------------------------*/
	if ( model_P->derateVIPct != 0 )
		for ( i = 0; i < numTablePoints; i++ )
			{
			model_P->pullup->VIs[i].i.min -= model_P->pullup->VIs[i].i.min * 
											 ( model_P->derateVIPct / 100 );
			model_P->pullup->VIs[i].i.max += model_P->pullup->VIs[i].i.max * 
											 ( model_P->derateVIPct / 100 );
			}
	
	}

if ( pulldownData_P != NULL )
	{
	/*------------------------------------------------------------------
	*	We're going to do almost the exact same thing for the pulldown
	*	data, so I'll leave off with the comments.
	*-----------------------------------------------------------------*/
	setupVoltages( CURVE_TYPE_PULLDOWN, model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );

	numInnerPoints = fabs( Vcc.typ - Gnd.typ + 2*diodeDrop ) / sweepStep + 1;
	pointsToSkip = (int) (fabs(sweepRange - (Vcc.typ - Gnd.typ + 2*diodeDrop))
				   / 2 * sweepStep * numLinearPoints );
	numTablePoints = 2 * numLinearPoints + numInnerPoints; 

	if ( create_VItable( &( model_P->pulldown ), numTablePoints ) )
		return( 1 );

	j = 0;
	for ( i = 0; i < numLinearPoints; i ++ )
		{
		model_P->pulldown->VIs[i] = pulldownData_P->VIs[j];
		j += pointsToSkip;
		}
	j -= pointsToSkip;
	currentMarker = i;

	while ( pulldownData_P->VIs[j].v != -diodeDrop )
		j++;
	
	for ( i = currentMarker; i < currentMarker + numInnerPoints;
		  i++ )
		{
		model_P->pulldown->VIs[i] = pulldownData_P->VIs[j];
		j++;
		}
	currentMarker = i;

	j--;
	for ( i = currentMarker; i < numTablePoints; i++ )
		{
		j += pointsToSkip;
		model_P->pulldown->VIs[i] = pulldownData_P->VIs[j];
		}
	if ( model_P->derateVIPct != 0 )
		for ( i = 0; i < numTablePoints; i++ )
			{
			model_P->pulldown->VIs[i].i.min -= model_P->pulldown->VIs[i].i.min
											   * ( model_P->derateVIPct / 100 );
			model_P->pulldown->VIs[i].i.max += model_P->pulldown->VIs[i].i.max 
											   * ( model_P->derateVIPct / 100 );
			}
	}

if ( powerClampData_P != NULL )
	{
	setupVoltages( CURVE_TYPE_POWER_CLAMP, model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );

	/*------------------------------------------------------------------
	*	All we need to do is read backwards through the table until we
	*	get to v < VccTyp, depositing the points into the model as we go.
	*-----------------------------------------------------------------*/
	numTablePoints = (int) ( fabs( sweepRange / sweepStep ) + 1 );
	if ( create_VItable( &( model_P->power_clamp ), numTablePoints ) )
		return( 1 );
	
	i = numTablePoints - 1;
	j = 0;
	while ( ( i >= 0 ) && ( powerClampData_P->VIs[i].v >= Vcc.typ ) )
		{
		model_P->power_clamp->VIs[j].v = Vcc.typ - powerClampData_P->VIs[i].v;
		model_P->power_clamp->VIs[j].i.typ = powerClampData_P->VIs[i].i.typ;
		model_P->power_clamp->VIs[j].i.min = powerClampData_P->VIs[i].i.min;
		model_P->power_clamp->VIs[j].i.max = powerClampData_P->VIs[i].i.max;
		i--;
		j++;
		}
	if ( model_P->derateVIPct != 0 )
		for ( i = 0; i < numTablePoints; i++ )
			{
			model_P->power_clamp->VIs[i].i.min -= 
						model_P->power_clamp->VIs[i].i.min * 
						( model_P->derateVIPct / 100 );
			model_P->power_clamp->VIs[i].i.max += 
						model_P->power_clamp->VIs[i].i.max * 
						( model_P->derateVIPct / 100 );
			}
	}

if ( gndClampData_P != NULL )
	{
	/*------------------------------------------------------------------
	*	Of course, a similar procedure applies for the ground clamp data.
	*-----------------------------------------------------------------*/
	setupVoltages( CURVE_TYPE_GND_CLAMP, model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );

	numTablePoints = (int) ( fabs( sweepRange / sweepStep ) + 1 );
	if ( create_VItable( &( model_P->gnd_clamp ), numTablePoints ) )
		return( 1 );
	
	j = 0;
	while ( ( j < numTablePoints ) && ( gndClampData_P->VIs[j].v <= Vcc.typ ) )
		{
		model_P->gnd_clamp->VIs[j] = gndClampData_P->VIs[j];
		j++;
		}
	if ( model_P->derateVIPct != 0 )
		for ( i = 0; i < numTablePoints; i++ )
			{
			model_P->gnd_clamp->VIs[i].i.min -= model_P->gnd_clamp->VIs[i].i.min
											* ( model_P->derateVIPct / 100 );
			model_P->gnd_clamp->VIs[i].i.max += model_P->gnd_clamp->VIs[i].i.max
											* ( model_P->derateVIPct / 100 );
			}
	}

return( 0 );
}


		



/* Function ***************************************************************
*
* NAME : analyzePin
*
* INPUTS :	ibisPin_Pt      currentPin_P    pointer to pin to be analyzed
*			ibisPin_Pt		enablePin_P		pointer to enable for current pin
*											(may be null)
*			ibisPin_Pt		inputPin_P		ptr to input for current pin
*											(may be null)
*			ibisPin_Pt		pullupPin_P		power pin ptr
*			ibisPin_Pt		pulldownPin_P 	gnd pin ptr
*			ibisPin_Pt		powerClampPin_P power pin ptr
*			ibisPin_Pt		gndClampPin_P	gnd pin ptr
*			int				spiceType		flavor of Spice
*			char*			spiceFile		filename of component spice file
*			char*			spiceCommand	spice invocation string
*			int				iterate			flag to skip processing if output
*											files exist already
*			int				cleanup			cleanup output files if no abort
*
*
* OUTPUTS : ibisPin_Pt 		currentPin_P -- pointer to pin struct
*
*
* RETURNS :     0 if successful
*               1 if unsuccessful
*
*
* NOTES / SIDE EFFECTS :  analyzePin is called only if the model type is
*				not one of POWER, GND or NC.  It determines
*				what curves (VI, VT) will be generated and
*				generates them.
*
*F*/

int 
analyzePin( INOUT 	ibisPin_Pt 		currentPin_P,	
			IN		ibisPin_Pt		enablePin_P,	
			IN		ibisPin_Pt		inputPin_P,		
			IN		ibisPin_Pt		pullupPin_P,	
			IN		ibisPin_Pt		pulldownPin_P,	
			IN		ibisPin_Pt		powerClampPin_P,
			IN		ibisPin_Pt		gndClampPin_P,	
			IN		int				spiceType,		
			IN		char			*spiceFile,		
			IN		char			*spiceCommand,	
			IN		int				iterate,		
			IN		int				cleanup )		
														

{

/*==== Declarations =====================================================*/
int					result = 0,		/* will be incremented on error */
					i = 0;
ibisVItable_Pt		pullupData_P = NULL,
					puDisabledData_P = NULL,
					pulldownData_P = NULL,
					pdDisabledData_P = NULL,
					powerClampData_P = NULL,
					gndClampData_P = NULL;
ibisWaveTable_Pt	risingWave_P = NULL,
					fallingWave_P = NULL;
ibisTypMinMax_t		Vcc,
					Gnd,
					sweepStart;
double				sweepRange = 0,
					sweepStep = 0,
					diodeDrop = 0;
int					index = 0;


/*==== Code =============================================================*/

if ( THIS_MODEL_NEEDS_PULLUP_DATA( currentPin_P->model_P->modelType ) )
	{
	setupVoltages( CURVE_TYPE_PULLUP, currentPin_P->model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );
	result += generateVICurve( &pullupData_P, currentPin_P, enablePin_P, 
							   inputPin_P, pullupPin_P, pulldownPin_P, 
							   Vcc, Gnd, sweepStart, sweepRange, sweepStep,
							   CURVE_TYPE_PULLUP, spiceType, spiceFile,
							   spiceCommand, ENABLE_OUTPUT, OUTPUT_HIGH,
							   iterate, cleanup );

	if ( enablePin_P != NULL )	/* we need to do a run with the thing... */
		{						/* ...disabled, to generate the diff curve */
		result  += generateVICurve( &puDisabledData_P, currentPin_P, 
									enablePin_P, inputPin_P, pullupPin_P,
									pulldownPin_P, Vcc, Gnd, sweepStart, 
									sweepRange, sweepStep, 
									CURVE_TYPE_DISABLED_PULLUP,
									spiceType, spiceFile, spiceCommand,
									DISABLE_OUTPUT, OUTPUT_HIGH, iterate,
									cleanup );

		if ( ( puDisabledData_P != NULL ) && ( pullupData_P != NULL ) )
			for ( i = 0; i < pullupData_P->size; i++ )
				{
				if ( ( IS_USE_NA( pullupData_P->VIs[i].i.typ ) ) ||
					 ( IS_USE_NA( puDisabledData_P->VIs[i].i.typ ) ) )
					pullupData_P->VIs[i].i.typ = USE_NA;
				else
					pullupData_P->VIs[i].i.typ -= 
											puDisabledData_P->VIs[i].i.typ;
				if ( ( IS_USE_NA( pullupData_P->VIs[i].i.min ) ) ||
					 ( IS_USE_NA( puDisabledData_P->VIs[i].i.min ) ) )
					pullupData_P->VIs[i].i.min = USE_NA;
				else
					pullupData_P->VIs[i].i.min -= 
											puDisabledData_P->VIs[i].i.min;
				if ( ( IS_USE_NA( pullupData_P->VIs[i].i.max ) ) ||
					 ( IS_USE_NA( puDisabledData_P->VIs[i].i.max ) ) )
					pullupData_P->VIs[i].i.max = USE_NA;
				else
					pullupData_P->VIs[i].i.max -= 
											puDisabledData_P->VIs[i].i.max;
				}
		else
			if ( pullupData_P != NULL )
				{
				fprintf( stderr, "s2ibis2: Cannot create pullup curve.  " );
				fprintf( stderr, "Deleting.\n");
				delete_VItable( &pullupData_P );
				}
		}
	}
	

if ( THIS_MODEL_NEEDS_PULLDOWN_DATA( currentPin_P->model_P->modelType ) )
	{
	setupVoltages( CURVE_TYPE_PULLDOWN, currentPin_P->model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );
	result += generateVICurve( &pulldownData_P, currentPin_P, enablePin_P, 
							   inputPin_P, pullupPin_P, pulldownPin_P, 
							   Vcc, Gnd, sweepStart, sweepRange, sweepStep,
							   CURVE_TYPE_PULLDOWN, spiceType, spiceFile,
							   spiceCommand, ENABLE_OUTPUT, OUTPUT_LOW,
							   iterate, cleanup );

	if ( enablePin_P != NULL )	/* we need to do a run with the thing... */
		{						/* ...disabled, to generate the diff curve */
		result  += generateVICurve( &pdDisabledData_P, currentPin_P, 
									enablePin_P, inputPin_P, pullupPin_P,
									pulldownPin_P, Vcc, Gnd, sweepStart,
									sweepRange, sweepStep, 
									CURVE_TYPE_DISABLED_PULLDOWN,
									spiceType, spiceFile, spiceCommand,
									DISABLE_OUTPUT, OUTPUT_LOW, iterate,
									cleanup );

		if ( ( pulldownData_P != NULL ) && ( pdDisabledData_P != NULL ) )
			for ( i = 0; i < pulldownData_P->size; i++ )
				{
				if ( ( IS_USE_NA( pulldownData_P->VIs[i].i.typ ) ) ||
					 ( IS_USE_NA( pdDisabledData_P->VIs[i].i.typ ) ) )
					pulldownData_P->VIs[i].i.typ = USE_NA;
				else
					pulldownData_P->VIs[i].i.typ -= 
											pdDisabledData_P->VIs[i].i.typ;
				if ( ( IS_USE_NA( pulldownData_P->VIs[i].i.min ) ) ||
					 ( IS_USE_NA( pdDisabledData_P->VIs[i].i.min ) ) )
					pulldownData_P->VIs[i].i.min = USE_NA;
				else
					pulldownData_P->VIs[i].i.min -= 
											pdDisabledData_P->VIs[i].i.min;
				if ( ( IS_USE_NA( pulldownData_P->VIs[i].i.max ) ) ||
					 ( IS_USE_NA( pdDisabledData_P->VIs[i].i.max ) ) )
					pulldownData_P->VIs[i].i.max = USE_NA;
				else
					pulldownData_P->VIs[i].i.max -= 
											pdDisabledData_P->VIs[i].i.max;
				}
		else
			if ( pulldownData_P != NULL )
				{
				fprintf( stderr, "s2ibis2: Cannot create pulldown curve.  " );
				fprintf( stderr, "Deleting.\n");
				delete_VItable( &pullupData_P );
				}
		}
	}
	


if ( THIS_MODEL_NEEDS_POWER_CLAMP_DATA( currentPin_P->model_P->modelType ) )
	{
	setupVoltages( CURVE_TYPE_POWER_CLAMP, currentPin_P->model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );
	result += generateVICurve( &powerClampData_P, currentPin_P, enablePin_P, 
							   inputPin_P, powerClampPin_P, gndClampPin_P, 
							   Vcc, Gnd, sweepStart, sweepRange, sweepStep,
							   CURVE_TYPE_POWER_CLAMP, spiceType, spiceFile,
							   spiceCommand, DISABLE_OUTPUT, OUTPUT_HIGH,
							   iterate, cleanup );
	}

if ( THIS_MODEL_NEEDS_GND_CLAMP_DATA( currentPin_P->model_P->modelType ) )
	{
	setupVoltages( CURVE_TYPE_GND_CLAMP, currentPin_P->model_P, &Vcc, &Gnd, 
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );
	result += generateVICurve( &gndClampData_P, currentPin_P, enablePin_P, 
							   inputPin_P, powerClampPin_P, gndClampPin_P, 
							   Vcc, Gnd, sweepStart, sweepRange, sweepStep,
							   CURVE_TYPE_GND_CLAMP, spiceType, spiceFile,
							   spiceCommand, DISABLE_OUTPUT, OUTPUT_LOW,
							   iterate, cleanup );
	}

sortVIData( currentPin_P->model_P, pullupData_P, pulldownData_P,
			powerClampData_P, gndClampData_P );

if ( pullupData_P != NULL )
	delete_VItable( &pullupData_P );
if ( pulldownData_P != NULL )
	delete_VItable( &pulldownData_P );
if ( powerClampData_P != NULL )
	delete_VItable( &powerClampData_P );
if ( gndClampData_P != NULL )
	delete_VItable( &gndClampData_P );


if ( THIS_MODEL_NEEDS_TRANSIENT_DATA( currentPin_P->model_P->modelType ) )
	{
	setupVoltages( CURVE_TYPE_RISING_RAMP, currentPin_P->model_P, &Vcc, &Gnd,
				   &sweepStart, &sweepRange, &sweepStep, &diodeDrop );
	result += generateRampData( currentPin_P, enablePin_P, inputPin_P,
								pullupPin_P, pulldownPin_P, Vcc, Gnd, 
								CURVE_TYPE_RISING_RAMP, spiceType, spiceFile, 
								spiceCommand, iterate, cleanup );
	result += generateRampData( currentPin_P, enablePin_P, inputPin_P,
								pullupPin_P, pulldownPin_P, Vcc, Gnd, 
								CURVE_TYPE_FALLING_RAMP, spiceType, spiceFile, 
								spiceCommand, iterate, cleanup );
	
	risingWave_P = currentPin_P->model_P->risingWave_P;
	fallingWave_P = currentPin_P->model_P->fallingWave_P;

	index = 0;
	while ( ( index < MAX_WAVEFORM_TABLES ) && ( risingWave_P != NULL ) )
		{
		result += generateWaveData( risingWave_P, currentPin_P, enablePin_P, 
									inputPin_P, pullupPin_P, pulldownPin_P, 
									Vcc, Gnd, CURVE_TYPE_RISING_WAVE, 
									spiceType, spiceFile, spiceCommand, 
									iterate, cleanup, index );
		index++;
		risingWave_P = risingWave_P->nextWaveTable_P;
		}
	
	index = 0;
	while ( ( index < MAX_WAVEFORM_TABLES ) && ( fallingWave_P != NULL ) )
		{
		result += generateWaveData( fallingWave_P, currentPin_P, enablePin_P, 
									inputPin_P, pullupPin_P, pulldownPin_P, 
									Vcc, Gnd, CURVE_TYPE_FALLING_WAVE, 
									spiceType, spiceFile, spiceCommand, 
									iterate, cleanup, index );
		index++;
		fallingWave_P = fallingWave_P->nextWaveTable_P;
		}
	}
	
return( result );
}



/* Function ***************************************************************
*
* NAME : analyzeComponent
*
* INPUTS :  ibisComponent_Pt component_P -- pointer to component to be analyzed
*			int spiceType				 --	flavor of Spice to run
*			int iterate					 --	flag to signal check for existing
*											Spice output files
*			int cleanup					 --	flag to signal cleanup of created
*											Spice input and output files
*			char *spiceCommand			 --	command string to invoke Spice
*
*
* OUTPUTS :     component_P -- pointer to component after analysis
*
*
* RETURNS :     0 if successful
*               1 if error 
*
*
* NOTES / SIDE EFFECTS :	analyzeComponent runs through each pin in the 
*				component structure.  It determines if the current pin model
*				is either an input or output pin (or terminator) i.e. not
*				one of POWER, GND or NC.  If it is, and the model for this
*				pin has not yet been generated, it calls analyzePin
*				to generate the correct model curves.
*F*/

int analyzeComponent(	INOUT	ibisComponent_Pt	component_P,
					 	IN		int					spiceType,
					 	IN		int					iterate,
					 	IN		int					cleanup,
					 	IN		char				*spiceCommand )

{
/*==== Declarations =====================================================*/
int				result = 0;
ibisPin_Pt		currentPin_P = NULL,
				pullupPin_P = NULL,
				pulldownPin_P = NULL,
				powerClampPin_P = NULL,
				gndClampPin_P = NULL,
				enablePin_P = NULL,
				inputPin_P = NULL;


/*==== Code =============================================================*/

if ( component_P->pins_P != NULL )
	{
	fprintf( stderr, "s2ibis2: Analyzing component %s.\n", 
			 component_P->component);
	for ( currentPin_P = component_P->pins_P;
	      currentPin_P != NULL;
	      currentPin_P = currentPin_P->nextPin_P)
		{
		/*------------------------------------------------------
		* First clear all the pointers that apply to each run.
		*-----------------------------------------------------*/
		pullupPin_P = NULL;
		pulldownPin_P = NULL;
		powerClampPin_P = NULL;
		gndClampPin_P = NULL;
		enablePin_P = NULL;
		inputPin_P = NULL;
			
		/*------------------------------------------------------
		* If this pin needs analysis and it has not yet been
		* analyzed, find its power, ground, enable and input pins, and call 
		* analyzePin to generate the appropriate model curves.
		*-----------------------------------------------------*/
		if ( ( THIS_PIN_NEEDS_ANALYSIS( currentPin_P->modelName ) ) && 
			!( currentPin_P->model_P->hasBeenAnalyzed ) )
			{
			result = findSupplyPins( &pullupPin_P, &pulldownPin_P, 
									 &powerClampPin_P, &gndClampPin_P,
									 currentPin_P, component_P->pins_P,
									 component_P->hasPinMapping ); 
			if ( !( result ) )
				{
				if ( strcmp( currentPin_P->enablePin, "" ) )
					{
					enablePin_P = getMatchingPin( currentPin_P->enablePin, 
												  component_P->pins_P );
					
					if ( enablePin_P == NULL )		/* if it's still NULL... */
						{							/* ...it wasn't in the... */
						fprintf( stderr,			/* ...pin list. */
			"         Couldn't find enable pin for pin %s.\n",
								 currentPin_P->pinName );
						fprintf( stderr, 
			"         Please make sure pin %s is in the pin list, or that\n",
								 currentPin_P->enablePin );
						fprintf( stderr,
			"         you have correctly spelled the enable pin name.\n" );
						result++;
						}
					}

				if ( strcmp( currentPin_P->inputPin, "" ) )
					{
					inputPin_P = getMatchingPin( currentPin_P->inputPin,
												 component_P->pins_P );
					if ( inputPin_P == NULL )		/* if it's still NULL... */
						{							/* ...it wasn't in the... */
						fprintf( stderr,			/* ...pin list. */
			"         Couldn't find input pin for pin %s.\n",
								 currentPin_P->pinName );
						fprintf( stderr, 
			"         Please make sure pin %s is in the pin list, or that\n",
								 currentPin_P->enablePin );
						fprintf( stderr,
			"         you have correctly spelled the input pin name.\n" );
						result++;
						}
					}

				if ( !( result ) )
					result = analyzePin( currentPin_P, enablePin_P, inputPin_P,
									 pullupPin_P, pulldownPin_P, 
									 powerClampPin_P, gndClampPin_P, spiceType, 
									 component_P->spiceFile, spiceCommand, 
									 iterate, cleanup );

				if ( result )
					{
					fprintf( stderr,"s2ibis2: Error in analysis ");
					fprintf( stderr, "for pin %s.\n", currentPin_P->pinName );
					}
				else
					currentPin_P->model_P->hasBeenAnalyzed++;
				}
			}
		}
	}
else
	{
	fprintf( stderr, "s2ibis2: Pin list for component %s is empty.\n",
			 component_P->component );
	fprintf( stderr, "         Please specify the pin list using the [Pin] " );
	fprintf( stderr, "keyword.\n" );
	result++;
	}

return( result );

}
