/* File *******************************************************************
*					North Carolina State University
*					Electronics Research Laboratory
*
* FILENAME :    s2ibis.y
*
* DESCRIPTION : This file provides the yacc (actually bison) parser which 
*				uses the lexer found in s2ibis.lex to parse s2ibis2 
*				command files.  While the s2ibis command language is 
*				very similar to the IBIS language, this lexer/parser 
*				combo will require some modification to correctly parse
*				IBIS files.
*
*
* FUNCTIONS:	yyparse (implicit)
*
*
* NOTES :
*
*
* REVISION HISTORY :
*       Date        Programmer:         Description
*		8/24/95		Alan Glaser			Beta release to some members of the
*					Michael Steer		IBIS community
*
*
*H*/
/*
 * s2ibis.y
 *
 *	YACC grammar, actions and C source for parsing an IBIS file.
 */
%{


extern int maimed;			/* TRUE if fatal errors encountered */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "s2ibis.h"
#include "s2iutil.h"


#define YYMAXDEPTH 500
#define IDENTIFIERDEPTH 10
#define MAX_TMP_STRING_LEN (MAX_IBIS_STRING_LENGTH+10)

#define RET(token) return(token);

#ifdef DEBUG
#undef RET(token)
#define RET(token)	 if(token!=eol)\
            printf("\nlineNum = %d\t%s \ttoken = %d",lineNum,yytext,token);\
		  else\
            printf("\nlineNum = %d\teol \ttoken = %d",lineNum - 1,token);\
		return(token);
#endif

#define COPY_YYTEXT  idIndex = (thisDepth++) % YYMAXDEPTH;\
		  strncpy(yylvalStr[idIndex], yytext, MIN(MAX_TOK_LEN,yyleng));


/*------------------------------------------------------------------------
*
*	The following globals are used by the lexer (s2ibis.lex).
*
*-----------------------------------------------------------------------*/
static char 		yylvalStr[YYMAXDEPTH][MAX_TOK_LEN + 1],	
										/* text of yylval.id */
					*lastNumberAsText,
					x_char_char;		/* new comment character */


/*------------------------------------------------------------------------
*
*	The following globals are used by the parser (s2ibis.y).
*
*-----------------------------------------------------------------------*/
char				string[MAX_TMP_STRING_LEN+1],
					*any_string;
static char			commentChar = '|',
	     			universalCommentIndicator = 127; /* backspace character */

int					length,
					stringLength,
					idIndex,			/* used to index tokens */
					thisDepth,
					lineNum,			/* line number in input file */
					ibisVersion;

double				globalFloat, 
					globalFloat1;

static  FILE		yin;				/* file pointer for input */

ibisTOP_Pt			ibis_P;

ibisGlobal_Pt		global_P;

ibisComponent_Pt	component_P, *component_PP;

ibisTypMinMax_Pt	typMinMax_P;

ibisPackage_Pt		package_P;

ibisModel_Pt		model_P, *model_PP;

ibisPin_Pt			pin_P, *pin_PP;

ibisPinMap_Pt		pinMap_P, *pinMap_PP;

ibisDiffPin_Pt		diffPin_P, *diffPin_PP;

ibisWaveTable_Pt	risingWave_P, *risingWave_PP, 
					fallingWave_P, *fallingWave_PP;

/* static functions */
void create_String(char *sin, char **sout);
void initialize_IBIS(IN int versionID);
void setUp_Component(void);
int setUp_FileName(void);
int setUp_SpiceFileName(void);
ibisPin_Pt getMatchingPin( char[], ibisPin_Pt );

%}


/*
 * YACC DEFINITIONS
 */

/*
 * The parser's stack can hold ints, doubles, and strings.
 */
%union {
	int ival;
	double dval;
	char *id;
	}



/* define token values */
%token COMPONENT_VOLTAGE_RANGE
%token COMPONENT_DERATE_RAMP
%token COMPONENT_POWER_CLAMP
%token COMPONENT_TEMP_RANGE
%token GLOBAL_VOLTAGE_RANGE
%token COMPONENT_DERATE_VI
%token COMPONENT_GND_CLAMP
%token COMPONENT_SIM_TIME
%token COMPONENT_PULLDOWN
%token GLOBAL_DERATE_RAMP
%token GLOBAL_POWER_CLAMP
%token GLOBAL_TEMP_RANGE
%token GLOBAL_DERATE_VI
%token GLOBAL_GND_CLAMP
%token COMPONENT_C_COMP
%token COMPONENT_PULLUP
%token COMPONENT_R_PKG
%token COMPONENT_L_PKG
%token COMPONENT_C_PKG
%token COMPONENT_RLOAD
%token GLOBAL_SIM_TIME
%token GLOBAL_PULLDOWN
%token I_O_OPEN_SOURCE
%token SCALED_FLOATLIT
%token I_O_OPEN_DRAIN
%token NEXT_COMPONENT  
%token GLOBAL_C_COMP
%token I_O_OPEN_SINK
%token NON_INVERTING
%token VOLTAGE_RANGE
%token PACKAGE_MODEL
%token GLOBAL_PULLUP
%token SPICE_COMMAND
%token IBIS_VERSION
%token FALLING_WAVE
%token COMMENT_CHAR
%token MANUFACTURER
%token GLOBAL_RLOAD
%token GLOBAL_R_PKG
%token GLOBAL_L_PKG
%token GLOBAL_C_PKG
%token VER_ONE_ONE
%token VER_TWO_ONE
%token RISING_WAVE
%token ACTIVE_HIGH
%token POWER_CLAMP
%token OPEN_SOURCE
%token DERATE_RAMP
%token OUTPUT_ECL
%token OPEN_DRAIN
%token TEMP_RANGE
%token DISCLAIMER
%token ACTIVE_LOW
%token MODEL_TYPE
%token WHITESPACE
%token TERMINATOR
%token SPICE_FILE
%token MODEL_FILE
%token SPICE_TYPE
%token CONT_LINE
%token DERATE_VI
%token OPEN_SINK
%token COMPONENT  
%token INVERTING
%token FILE_NAME
%token COPYRIGHT
%token GND_CLAMP
%token INPUT_ECL
%token TRI_STATE
%token DIFF_PIN
%token PULLDOWN
%token POLARITY
%token FILE_REV
%token SIM_TIME
%token ITERATE
%token PIN_MAP
%token PACKAGE
%token DV_DT_R
%token DV_DT_F
%token I_O_ECL
%token SPECTRE
%token CLEANUP
%token ENABLE
%token SOURCE
%token PULLUP
%token C_COMP
%token RPOWER
%token END_kw
%token OUTPUT
%token PSPICE
%token HSPICE
%token SPICE2
%token SPICE3
%token R_PIN
%token L_PIN
%token C_PIN
%token POWER
%token NOTES
%token RLOAD
%token VMEAS
%token MODEL
%token INPUT
%token VINL
%token VINH
%token DATE
%token RAMP
%token OVER
%token CREF
%token RREF
%token VREF
%token RGND
%token I_O
%token GND
%token RAC
%token CAC
%token PIN
%token eol
%token eof
%token NC
%token NA



/* Entities */

%token <id>   IDENTIFIER	/* identifier, not a keyword    */
%token <dval> FLOATLIT		/* floating-pt literal		    */
%token <ival> INTLIT		/* integer literal			    */


/* ASCII chars are their own tokens */


%start	begin_parse
/*
 * THE PARSER
 */

%%

/*-------------------------------------------------------------------------
*
*	The first thing the parser wants to see is the version number.
*
*------------------------------------------------------------------------*/
begin_parse	: eols IBIS_VERSION VER_TWO_ONE
		  { initialize_IBIS( VERSION_TWO_ONE ); }
		  eols input_2.1 eof { return( 0 ); }
		| eols IBIS_VERSION VER_ONE_ONE
			{ initialize_IBIS( VERSION_ONE_ONE ); }
		  eols input_2.1 eof { return( 0 ); }
		| error { 
			fprintf( stderr, "s2ibis: Please begin s2ibis command " );
			fprintf( stderr, "file with [IBIS Ver] command.\n" );
			maimed = 1; 
			return(1); 
			}
		;

/*-------------------------------------------------------------------------
*
*	Next we look for the "global" defines.
*
*------------------------------------------------------------------------*/
input_2.1	:
		| input_2.1 COMMENT_CHAR any_identifier eols
			{
			commentChar = any_string[0];
			global_P->commentChar = any_string[0];
			}
		| input_2.1 error { maimed = 1; return(1); }
		| input_2.1 CLEANUP eols
			{ ibis_P->cleanup = 1; }
		| input_2.1 ITERATE eols
			{ ibis_P->iterate = 1; }
		| input_2.1 FILE_NAME any_identifier
			{ setUp_FileName(); }
		  eols
		| input_2.1 SPICE_TYPE spice_type
		| input_2.1 FILE_REV any_identifier
		    { LOAD_STRING( ibis_P->fileRev, any_string, MAX_FILE_REV_LENGTH ); }
		  eols
		| input_2.1 DATE saveString
			{ LOAD_STRING( ibis_P->date, string, MAX_DATE_STRING_LENGTH); }
		| input_2.1 SPICE_COMMAND saveString
			{ create_String( string, &( ibis_P->spiceCommand ) ); }
		| input_2.1 SOURCE saveString
			{ create_String(string, &(ibis_P->source)); }
		| input_2.1 NOTES saveString
			{ create_String(string, &(ibis_P->notes)); }
		| input_2.1 DISCLAIMER saveString
			{ create_String(string, &(ibis_P->disclaimer)); }
		| input_2.1 COPYRIGHT saveString
			{ create_String(string, &(ibis_P->copyright)); }
		| input_2.1 GLOBAL_TEMP_RANGE
			{ typMinMax_P = &( global_P->tempRange ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_VOLTAGE_RANGE
			{ typMinMax_P = &( global_P->voltageRange ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_PULLUP
			{ typMinMax_P = &( global_P->pullupRef ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_PULLDOWN
			{ typMinMax_P = &( global_P->pulldownRef ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_POWER_CLAMP
			{ typMinMax_P = &( global_P->powerClampRef ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_GND_CLAMP
			{ typMinMax_P = &( global_P->gndClampRef ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_RLOAD number_as_double eols
			{ global_P->Rload = globalFloat; }
		| input_2.1 GLOBAL_DERATE_VI number_as_double eols
			{ global_P->derateVIPct = globalFloat; }
		| input_2.1 GLOBAL_DERATE_RAMP number_as_double eols
			{ global_P->derateRampPct = globalFloat; }
		| input_2.1 GLOBAL_SIM_TIME number_as_double eols
			{ global_P->simTime = globalFloat; }
		| input_2.1 GLOBAL_C_COMP
			{ typMinMax_P = &( global_P->c_comp ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_R_PKG
			{ typMinMax_P = &( global_P->pinParasitics.R_pkg ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_L_PKG
			{ typMinMax_P = &( global_P->pinParasitics.L_pkg ); }
		  typ_min_max eols
		| input_2.1 GLOBAL_C_PKG
			{ typMinMax_P = &( global_P->pinParasitics.C_pkg ); }
		  typ_min_max eols
		| input_2.1  component_1.0
		;

/*-------------------------------------------------------------------------
*
*	This handles the spice type.
*
*------------------------------------------------------------------------*/
spice_type:	HSPICE eols { ibis_P->spiceType = SPICE_TYPE_HSPICE; }
		| PSPICE eols { ibis_P->spiceType = SPICE_TYPE_PSPICE; }
		| SPICE2 eols { ibis_P->spiceType = SPICE_TYPE_SPICE2; }
		| SPICE3 eols { ibis_P->spiceType = SPICE_TYPE_SPICE3; }
		| SPECTRE eols { ibis_P->spiceType = SPICE_TYPE_SPECTRE; }
		;


/*-------------------------------------------------------------------------
*
*	This sets up the component and all its parameters.
*
*------------------------------------------------------------------------*/
component_1.0	:  COMPONENT saveString
			{ setUp_Component(); }
		   body_of_component_1.0
		;


body_of_component_1.0 :
	    |  body_of_component_1.0 MANUFACTURER saveString
			{
	     	LOAD_STRING( component_P->manufacturer, string,
	     	 	  		 MAX_MANUFACTURERS_NAME_LENGTH);
			}
		| body_of_component_1.0 SPICE_FILE any_identifier
			{ setUp_SpiceFileName(); }
		  eols
	 	| body_of_component_1.0 PACKAGE_MODEL saveString
			{
			LOAD_STRING( component_P->packageModel, string,
						 MAX_PACKAGE_MODEL_NAME_LENGTH );
			}
		| body_of_component_1.0 COMPONENT_TEMP_RANGE
			{ typMinMax_P = &( component_P->tempRange ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_VOLTAGE_RANGE
			{ typMinMax_P = &( component_P->voltageRange ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_PULLUP
			{ typMinMax_P = &( component_P->pullupRef ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_PULLDOWN
			{ typMinMax_P = &( component_P->pulldownRef ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_POWER_CLAMP
			{ typMinMax_P = &( component_P->powerClampRef ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_GND_CLAMP
			{ typMinMax_P = &( component_P->gndClampRef ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_RLOAD number_as_double eols
			{ component_P->Rload = globalFloat; }
		| body_of_component_1.0 COMPONENT_DERATE_VI number_as_double eols
			{ component_P->derateVIPct = globalFloat; }
		| body_of_component_1.0 COMPONENT_DERATE_RAMP number_as_double eols
			{ component_P->derateRampPct = globalFloat; }
		| body_of_component_1.0 COMPONENT_SIM_TIME number_as_double eols
			{ component_P->simTime = globalFloat; }
		| body_of_component_1.0 COMPONENT_C_COMP
			{ typMinMax_P = &( component_P->c_comp ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_R_PKG
			{ typMinMax_P = &( component_P->pinParasitics.R_pkg ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_L_PKG
			{ typMinMax_P = &( component_P->pinParasitics.L_pkg ); }
		  typ_min_max eols
		| body_of_component_1.0 COMPONENT_C_PKG
			{ typMinMax_P = &( component_P->pinParasitics.C_pkg ); }
		  typ_min_max eols
		| body_of_component_1.0 DIFF_PIN eols
		  body_of_diff_pin
		| body_of_component_1.0 PIN_MAP eols
			{ component_P->hasPinMapping = 1; }
		  body_of_pin_map
		| body_of_component_1.0 PIN eols
		  body_of_pin
		| body_of_component_1.0 MODEL any_identifier
			{
		        create_Model(model_PP, &model_P);
		        model_PP = &model_P;
				LOAD_STRING( model_P->modelName, any_string, 
							 MAX_MODEL_NAME_LENGTH )
				risingWave_PP = &( model_P->risingWave_P );
				fallingWave_PP = &( model_P->fallingWave_P );
			}
		  eols
		  body_of_model_2.0
		;



/*-------------------------------------------------------------------------
*
*	The diff pin list consists of two types of lines: 
*
*	A)	pin name, inv pin, vdiff, tdelay typ
*
*	B)	pin name, inv pin, vdiff, tdelay typ, tdelay min, tdelay max
*
*------------------------------------------------------------------------*/
body_of_diff_pin: 
		| body_of_diff_pin diff_pin_line eols
		;

diff_pin_line: any_identifier
			{
			createDiffPin( diffPin_PP, &diffPin_P );
			diffPin_PP = &diffPin_P;
			LOAD_STRING( diffPin_P->pinName, any_string, MAX_PIN_NAME_LENGTH );
			}
		  any_identifier
		  	{ LOAD_STRING( diffPin_P->invPinName, any_string, 
						   MAX_PIN_NAME_LENGTH ); }
		  number_or_NA
		  	{ diffPin_P->vdiff = globalFloat; }
		  number_or_NA
		  	{ diffPin_P->tdelayTyp = globalFloat; }
		  maybe_tdelay_min_and_max
		;

maybe_tdelay_min_and_max: { diffPin_P->tdelayMin = NOT_USED;
							diffPin_P->tdelayMax = NOT_USED; }
		| number_or_NA
			{ diffPin_P->tdelayMin = globalFloat; }
		  number_or_NA
		  	{ diffPin_P->tdelayMax = globalFloat; }
		;


/*-------------------------------------------------------------------------
*
*	The pin map consists of two types of lines: 
*
*	A)	pin name, pullup ref, pulldown ref
*
*	B)	pin name, pullup ref, pulldown ref, power clamp ref, ground clamp ref
*
*------------------------------------------------------------------------*/
body_of_pin_map :
		| body_of_pin_map pin_map_line_type1 eols
		;

pin_map_line_type1 : any_identifier
			{
			createPinMap( pinMap_PP, &pinMap_P );
			pinMap_PP = &pinMap_P;
			LOAD_STRING( pinMap_P->pinName, any_string,
						 MAX_PIN_MAPPING_NAME_LENGTH );
			}
		  any_identifier
		  	{ LOAD_STRING( pinMap_P->pullupRef, any_string,
						   MAX_PIN_MAPPING_NAME_LENGTH ); }
		  any_identifier
		  	{ LOAD_STRING( pinMap_P->pulldownRef, any_string,
						   MAX_PIN_MAPPING_NAME_LENGTH ); }
		  maybe_clamp_refs
		;

maybe_clamp_refs : 
		| any_identifier
		  	{ LOAD_STRING( pinMap_P->powerClampRef, any_string,
						   MAX_PIN_MAPPING_NAME_LENGTH ); }
		  any_identifier
		  	{ LOAD_STRING( pinMap_P->gndClampRef, any_string,
						   MAX_PIN_MAPPING_NAME_LENGTH ); }
		;



/*-------------------------------------------------------------------------
*
* We need to be able to handle six different cases of pin data lines:
*
*	A) input pin, no parasitics:
*		pin_name spice_node signal_name model_name
*
*	B) input pin with parasitics:
*		pin_name spice_node signal_name model_name R_pin L_pin C_pin
*
*	C) Output pin, no parasitics:
*		pin_name spice_node signal_name model_name sim_time
*		input_pin_name
*
*	D) Output pin with parasitics:
*		pin_name spice_node signal_name model_name sim_time R_pin L_pin C_pin
*		input_pin_name
*
*	E) Output pin with enable, no parasitics:
*		pin_name spice_node signal_name model_name sim_time
*		input_pin_name enable_pin_name
*
*	F) Output pin with enable and parasitics:
*		pin_name spice_node signal_name model_name sim_time R_pin L_pin C_pin
*		input_pin_name enable_pin_name
*
*------------------------------------------------------------------------*/
body_of_pin :	
		| body_of_pin pin_line1 eols
		| body_of_pin pin_line1 eol CONT_LINE pin_line2_input_and_enable eols
		| body_of_pin pin_line1 eol CONT_LINE pin_line2_input_only eols
		;

pin_line1: any_identifier
			{
			create_Pin( pin_PP, &pin_P );
			pin_PP = &pin_P;
			LOAD_STRING( pin_P->pinName, any_string, MAX_PIN_NAME_LENGTH );
			}
		  any_identifier
		  	{
			LOAD_STRING( pin_P->spiceNodeName, any_string, 
						 MAX_PIN_NAME_LENGTH );
			}
		  any_identifier
		  	{
			LOAD_STRING( pin_P->signalName, any_string, 
						 MAX_SIGNAL_NAME_LENGTH );
			}
		  any_identifier
		  	{
			LOAD_STRING( pin_P->modelName, any_string, MAX_MODEL_NAME_LENGTH );
			}
		  maybe_pin_rlc
		;

pin_line2_input_only:	any_identifier
		  	{
			LOAD_STRING( pin_P->inputPin, any_string, MAX_PIN_NAME_LENGTH );
			}
		;

pin_line2_input_and_enable:	any_identifier
		  	{
			LOAD_STRING( pin_P->inputPin, any_string, MAX_PIN_NAME_LENGTH );
			}
		  any_identifier
		  	{
			LOAD_STRING( pin_P->enablePin, any_string, MAX_PIN_NAME_LENGTH );
			}
		;


/*-------------------------------------------------------------------------
*
*	This handles the option of pin parasitics.
*
*------------------------------------------------------------------------*/
maybe_pin_rlc	:
		| number_or_NA	{ pin_P->R_pin = globalFloat;}
		  number_or_NA	{ pin_P->L_pin = globalFloat;}
		  number_or_NA	{ pin_P->C_pin = globalFloat;}
		;


/*-------------------------------------------------------------------------
*
*	Now we handle setting up the model.
*
*------------------------------------------------------------------------*/
body_of_model_2.0 : 
		| body_of_model_2.0 MODEL_TYPE model_type_2.0 eols
		| body_of_model_2.0 MODEL_FILE model_file eols
		| body_of_model_2.0 polarity_1.0
		| body_of_model_2.0 enable_1.0
		| body_of_model_2.0 SIM_TIME number_as_double eols
			{
			model_P->simTime = globalFloat;
			}
		| body_of_model_2.0 VINL number_as_double eols
			{ 
			model_P->Vinl = globalFloat; 
			}
		| body_of_model_2.0 VINH number_as_double eols
			{ 
			model_P->Vinh = globalFloat; 
			}
		| body_of_model_2.0 VMEAS number_as_double eols
			{ 
			model_P->Vmeas = globalFloat; 
			}
		| body_of_model_2.0 CREF number_as_double eols
			{ 
			model_P->Cref = globalFloat; 
			}
		| body_of_model_2.0 RREF number_as_double eols
			{ 
			model_P->Rref = globalFloat; 
			}
		| body_of_model_2.0 VREF number_as_double eols
			{ 
			model_P->Vref = globalFloat; 
			}
		| body_of_model_2.0 RLOAD number_as_double eols
			{ 
			model_P->Rload = globalFloat; 
			}
		| body_of_model_2.0 DERATE_RAMP number_as_double eols
			{ 
			model_P->derateRampPct = globalFloat; 
			}
		| body_of_model_2.0 DERATE_VI number_as_double eols
			{ 
			model_P->derateVIPct = globalFloat; 
			}
		| body_of_model_2.0 C_COMP
			{ 
			typMinMax_P = &(model_P->c_comp); 
			}
		  typ_min_max eols
		| body_of_model_2.0 TEMP_RANGE
			{ 
			typMinMax_P = &(model_P->tempRange); 
			}
		  typ_min_max eols
		| body_of_model_2.0 VOLTAGE_RANGE
			{ 
			typMinMax_P = &(model_P->voltageRange); 
			}
		  typ_min_max eols
		| body_of_model_2.0 PULLUP
			{ 
			typMinMax_P = &(model_P->pullupRef); 
			}
		  typ_min_max eols
		| body_of_model_2.0 PULLDOWN
			{ 
			typMinMax_P = &(model_P->pulldownRef); 
			}
		  typ_min_max eols
		| body_of_model_2.0 POWER_CLAMP
			{ 
			typMinMax_P = &(model_P->powerClampRef); 
			}
		  typ_min_max eols
		| body_of_model_2.0 GND_CLAMP
			{ 
			typMinMax_P = &(model_P->gndClampRef); 
			}
		  typ_min_max eols
		| body_of_model_2.0 RGND
			{ 
			typMinMax_P = &(model_P->Rgnd); 
			}
		  typ_min_max eols
		| body_of_model_2.0 RPOWER
			{ 
			typMinMax_P = &(model_P->Rpower); 
			}
		  typ_min_max eols
		| body_of_model_2.0 RAC
			{ 
			typMinMax_P = &(model_P->Rac); 
			}
		  typ_min_max eols
		| body_of_model_2.0 CAC
			{ 
			typMinMax_P = &(model_P->Cac); 
			}
		  typ_min_max eols
		| body_of_model_2.0 RISING_WAVE 
			{
			createWaveTable( risingWave_PP, 1 );
			risingWave_P = *risingWave_PP;
			risingWave_PP = &( risingWave_P->nextWaveTable_P );
			}
			body_of_rising_wave
		| body_of_model_2.0 FALLING_WAVE
			{
			createWaveTable( fallingWave_PP, 1 );
			fallingWave_P = *fallingWave_PP;
			fallingWave_PP = &( fallingWave_P->nextWaveTable_P );
			} 
			body_of_falling_wave
		;



model_type_2.0	: INPUT
		  { model_P->modelType = MODEL_TYPE_INPUT; }
		| OUTPUT
		  { model_P->modelType = MODEL_TYPE_OUTPUT; }
		| TRI_STATE
		  { model_P->modelType = MODEL_TYPE_3_STATE; }
		| OPEN_DRAIN
		  { model_P->modelType = MODEL_TYPE_OPEN_DRAIN; }
		| I_O
		  { model_P->modelType = MODEL_TYPE_I_O; }
		| I_O_OPEN_DRAIN
		  { model_P->modelType = MODEL_TYPE_I_O_OPEN_DRAIN; }
		| I_O_OPEN_SINK
		  { model_P->modelType = MODEL_TYPE_I_O_OPEN_SINK; }
		| I_O_OPEN_SOURCE
		  { model_P->modelType = MODEL_TYPE_I_O_OPEN_SOURCE; }
		| INPUT_ECL
		  { model_P->modelType = MODEL_TYPE_INPUT_ECL; }
		| I_O_ECL
		  { model_P->modelType = MODEL_TYPE_I_O_ECL; }
		| TERMINATOR
		  { model_P->modelType = MODEL_TYPE_TERMINATOR; }
		| OPEN_SINK
		  { model_P->modelType = MODEL_TYPE_OPEN_SINK; }
		| OPEN_SOURCE
		  { model_P->modelType = MODEL_TYPE_OPEN_SOURCE; }
		| OUTPUT_ECL
		  { model_P->modelType = MODEL_TYPE_OUTPUT_ECL; }
		; 

polarity_1.0	: POLARITY NON_INVERTING eols
		  { model_P->polarity = MODEL_POLARITY_NON_INVERTING; }
		| POLARITY INVERTING eols
		  { model_P->polarity = MODEL_POLARITY_INVERTING; }
		;

enable_1.0	: ENABLE ACTIVE_HIGH eols
		  { model_P->enable = MODEL_ENABLE_ACTIVE_HIGH; }
		| ENABLE ACTIVE_LOW  eols
		  { model_P->enable = MODEL_ENABLE_ACTIVE_LOW; }
		;


body_of_rising_wave: number_as_double { risingWave_P->R_fixture = globalFloat; }
		  number_as_double { risingWave_P->V_fixture = globalFloat; }
		  number_or_NA { risingWave_P->V_fixture_min = globalFloat; }
		  number_or_NA { risingWave_P->V_fixture_max = globalFloat; }
		  number_or_NA { risingWave_P->L_fixture = globalFloat; }
		  number_or_NA { risingWave_P->C_fixture = globalFloat; }
		  number_or_NA { risingWave_P->R_dut = globalFloat; }
		  number_or_NA { risingWave_P->L_dut = globalFloat; }
		  number_or_NA { risingWave_P->C_dut = globalFloat; }
		  eols
		;

body_of_falling_wave: number_as_double { fallingWave_P->R_fixture = globalFloat; }
		  number_as_double { fallingWave_P->V_fixture = globalFloat; }
		  number_or_NA { fallingWave_P->V_fixture_min = globalFloat; }
		  number_or_NA { fallingWave_P->V_fixture_max = globalFloat; }
		  number_or_NA { fallingWave_P->L_fixture = globalFloat; }
		  number_or_NA { fallingWave_P->C_fixture = globalFloat; }
		  number_or_NA { fallingWave_P->R_dut = globalFloat; }
		  number_or_NA { fallingWave_P->L_dut = globalFloat; }
		  number_or_NA { fallingWave_P->C_dut = globalFloat; }
		  eols
		;

model_file: any_identifier
		  	{
			LOAD_STRING( model_P->modelFileTyp, any_string, 
						 MAX_FILENAME_LENGTH );
			}
		  any_identifier
		  	{
			LOAD_STRING( model_P->modelFileMin, any_string, 
						 MAX_FILENAME_LENGTH );
			}
		  any_identifier
		  	{
			LOAD_STRING( model_P->modelFileMax, any_string, 
						 MAX_FILENAME_LENGTH );
			}
		;

/*
 * Generic subsequences
 */
eols		: eol	/* Skip blank lines and lines with ... */
		| eols eol	/* ... comments only */
		;

any_identifier	: IDENTIFIER	{ any_string= $<id>1; }
		| INTLIT	{ any_string= lastNumberAsText; }
		| FLOATLIT	{ any_string= lastNumberAsText; }
		| GND		{ strcpy( any_string, "GND" ); }
		| POWER		{ strcpy( any_string, "POWER" ); }
		| NC		{ strcpy( any_string, "NC" ); }
		| NA		{ strcpy( any_string, "NA" ); }
		;

typ_min_max	: number_or_NA { typMinMax_P->typ = globalFloat; }
		  number_or_NA { typMinMax_P->min = globalFloat; }
		  number_or_NA { typMinMax_P->max = globalFloat; }
		;

number_or_NA	: number_as_double
		| NA {globalFloat = USE_NA;}
		;

number_as_double :	FLOATLIT { globalFloat = $<dval>1; }
		|	INTLIT	 { globalFloat = (double) $<ival>1; }
		;


saveString	: { stringLength = 0; string[0] = 0; }
		  any_identifier_eol
		  saveString1 { string[stringLength] = 0;/* replace last eol */}
		;

saveString1	:
		| saveString1 any_identifier_eol
		;

any_identifier_eol : any_identifier
			{
			length = MAX(MAX_TMP_STRING_LEN-stringLength,0);
			if(length)
			  {
			  strncat(string, any_string,length);
			  strncat(string,  " ", length--);
			  stringLength = strlen(string);
			  }
			}
		| eol
			{
			if(!stringLength)
			  {
			  length = MAX(MAX_TMP_STRING_LEN-stringLength,0);
			  strncat(string,  " ", length);
			  stringLength++;
			  }
			}
		;

%%

#include "s2ilex.c"



/*
 * yyerror--
 *	Do something (print it for now) with an error message.
 */
int yyerror(char *s)

{

char *b, *c;
int i;

fprintf( stderr, "\n\ns2ibis2: Error at line %d: %s\n", lineNum, s );
fprintf( stderr, "Token:\n" );
for ( c = yytext; (*c && (*c != '\n')); c++ )
	if ( *c == universalCommentIndicator )
		putchar( commentChar );
	else
		putchar( *c );
fprintf( stderr, "\n");
#ifdef FLEX_SCANNER
	fprintf( stderr, "Current Buffer:\n");
	for (b = yytext;
	  (b >= YY_CURRENT_BUFFER->yy_ch_buf && (*b != '\n')); b--);
	for (c = b+1,i=0; (*c && (*c != '\n') && i < MAX_LINE_LENGTH); i++, c++)
	  if (*c == universalCommentIndicator)
	    putc(commentChar, stderr);
	  else
	    putc(*c, stderr);
	fprintf(stderr, "\n");
	for ( c = b+1, i=0; 
		  strncmp( c,yytext,yyleng ) && *c && (*c != '\n') && 
		  i < MAX_LINE_LENGTH;
		  i++, c++ ) 
		{
		if (*c == '\t')
			putc('\t', stderr);
		else
			putc(' ', stderr);
		}
	putc('^', stderr);
	fprintf( stderr, "\n\n");
#endif

maimed = 1;
return(1);
}


void InitScan(char *file)
{
if ( (yyin = fopen( file, "r" ) ) == NULL )
	{
	fprintf( stderr, "s2ibis2: unable to open input file for reading.\n" );
	exit( 1 );
	}
else
	{
	lineNum = 1;
	thisDepth = 0;
	maimed = 0;
	}
}


/*
 * s2ibisRead
 *
 * This is the entry point to the parser
 *
 * Parameters:
 *	file: Name of the file to be searched
 *	ibisStruc_PP: pointer to the pointer of the data structure that
 *			  contains the information in the IBIS file>
 *	instruction: Instruction passed from the calling routin to
 *			  ibis1_1Read. It must be one of the following:
 *			  "new"		create a new data structure and
 *					transfer all of the information
 *					in the IBIS file into it.
 */

int s2ibisRead(	IN 		char *file, 
				INOUT	ibisTOP_Pt *ibisStruc_PP,
				INOUT	ibisGlobal_Pt	*ibisGlobal_PP,
               	IN 		char *instruction, 
				OUT 	char **report_P)
{
InitScan( file );
if( !strcmp( instruction,"new" ) )
	{
	create_TOP( ibisStruc_PP, file );
	createGlobal( ibisGlobal_PP );
	}
else
	{
	fprintf( stderr, "s2ibis2: Internal error--" );
	fprintf( stderr, "illegal instruction to parser.\n" );
	fprintf( stderr, "s2ibis2: Quit mucking about with my source code!\n" );
	return( 1 );
	}
    
ibis_P = *ibisStruc_PP;
global_P = *ibisGlobal_PP;

fprintf( stderr, "s2ibis2: Reading input file %s...", file );
if ( maimed || yyparse() )
	{
	fprintf( stderr, "s2ibis2: Errors in IBIS file prevent continuation.\n" );
	fclose( yyin );
	return( 1 );
	}
fclose( yyin );
fprintf( stderr, "done.\n" );
return( 0 );
}





void create_String(char *sin, char **sout)
{
int i = 0;

while ( ( sin[0] == ' ' ) || ( sin[0] == '\t' ) )
	sin++;
length = strlen(sin);
if ( length > MAX_IBIS_STRING_LENGTH )
	{
	sin[MAX_IBIS_STRING_LENGTH] = ' ';
	sin[MAX_IBIS_STRING_LENGTH+1] = '.';
	sin[MAX_IBIS_STRING_LENGTH+2] = '.';
	sin[MAX_IBIS_STRING_LENGTH+3] = '.';
	sin[MAX_IBIS_STRING_LENGTH+4] = 0;
	length = MAX_IBIS_STRING_LENGTH+4;
	}
if ( (*sout = (char *) MALLOC(length +1)) == NULL )
	yyerror("Out of memory");
else
	strncpy(*sout,sin,length);	
}

void recreate_String(char *sin, char **sout)
{
if(*sout != NULL) 
	FREE(*sout);
create_String(sin, sout);
}

/*
 * initialize_IBIS
 */
void initialize_IBIS(IN int versionID)
{
ibis_P->ibisVersion = versionID;
component_PP = &( ibis_P->components_P );
component_P = NULL;
model_P = NULL;
model_PP = NULL;
pin_P = NULL;
pin_PP = NULL;
pinMap_P = NULL;
pinMap_PP = NULL;
diffPin_P = NULL;
diffPin_PP = NULL;
risingWave_P = NULL;
risingWave_PP = NULL;
fallingWave_P = NULL;
fallingWave_PP = NULL;
return;
}

/*
 * setUp_Component();
 */
void setUp_Component()
{
create_Component( component_PP, &component_P );
component_PP = &component_P;
LOAD_STRING( component_P->component, string, MAX_COMPONENT_NAME_LENGTH );
pin_PP = &( component_P->pins_P );
model_PP = &( component_P->models_P );
pinMap_PP = &( component_P->pinMap_P );
diffPin_PP = &( component_P->diffPin_P );
}



/*
 * setUp_FileName;
 */
int setUp_FileName( void )

{
int	i,
	length;

length = strlen( any_string );
INITIAL_LOAD_STRING( ibis_P->thisFileName, any_string, MAX_FILENAME_LENGTH );
if ( ( length > MAX_FILENAME_LENGTH ) ||
     !strcmp( &((ibis_P->thisFileName)[length-3]), ".ibs" ) )
	{
	yyerror( "Invalid filename: must be of the form 'xxxxxxxx.ibs'" );
	maimed = 1;
	return( 1 );
	}
for ( i = 0; i < length - 4; i++ )
	if ( isupper( any_string[i] ) )
		{
		yyerror( "Invalid filename: must be lower case" );
		maimed = 1;
		return( 1 );
		}
}




/*
 * setUp_SpiceFileName;
 */
int setUp_SpiceFileName( void )

{
int	i,
	length;

length = strlen( any_string );
INITIAL_LOAD_STRING( component_P->spiceFile, any_string, MAX_FILENAME_LENGTH );
if ( ( length > MAX_FILENAME_LENGTH ) )
	{
	yyerror( "Invalid filename length: must be of the form 'xxxxxxxx.xxx'" );
	maimed = 1;
	return( 1 );
	}
for ( i = 0; i < length - 4; i++ )
	if ( isupper( any_string[i] ) )
		{
		yyerror( "Invalid filename: must be lower case" );
		maimed = 1;
		return( 1 );
		}
}
