/* File *******************************************************************
*					North Carolina State University
*					Electronics Research Laboratory
*
* FILENAME :	s2ispice.c
*
*
* DESCRIPTION :	Contains all of the Spice interface code for s2ibis2.
*
*
* FUNCTIONS :	setupPowerAndTempCommands
*				setupDCSweepCommands
*				setupTranCommands
*				setupSpiceFileNames
*				setupSpiceInputFile
*				setPinTran
*				setPinDC
*				callSpice
*				checkForAbort
*				getSpiceVIData
*				generateVICurve
*				getSpiceRampData
*				generateRampData
*				binTranData
*				getSpiceWaveData
*				generateWaveData
*
*
* $Id: s2ispice.c,v 0.91 1995/09/20 21:47:06 awglaser Exp $
*
*
* $Log: s2ispice.c,v $
*   Revision 0.91  1995/09/20  21:47:06  awglaser
*   -- Added standard file header with Id, Log and copyright info.
*   -- changed some fscanfs to fgets to deal with problem reading
*      HSpice data files (still investigating why problem occurs, but
*      fgets instead of fscanf seems to get rid of the problem).
*
*      Bug report by Deepak Savadatti.
*
*
*
* Copyright 1994, North Carolina State University.
*
* North Carolina State University (NCSU) and the author hereby disclaim
* all implied warranties, including the implied warranties of merchant-
* ability and fitness for a particular purpose. NCSU is not liable for
* any damages incurred by the use of this software and/or documentation,
* including direct, indirect, special, incidental, or consequential
* damages.
*
* NCSU grants the bearer the right to modify, copy, and redistribute
* this software and documentation, but requests that NCSU be acknowledged
* for its contribution in the resulting product.
*
*H*/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include "s2ibis.h"
#include "s2istrng.h"

void
setupPowerAndTempCommands(	OUT	char		*powerBuffer,
							OUT	char		*temperatureBuffer,
							IN	int			spiceType,
							IN	ibisPin_Pt	powerPin_P,
							IN	ibisPin_Pt	gndPin_P,
							IN	double		Vcc,
							IN	double		Gnd,
							IN	double		temperature )

{
/*==== Declarations =====================================================*/
char	dummyBuffer[128];			/* temporary buffer */

/*==== Code =============================================================*/
if ( powerPin_P != NULL )
	sprintf( powerBuffer, "VCCS2I %s 0 DC %g\n", powerPin_P->spiceNodeName, 
			 Vcc );
if ( gndPin_P != NULL )
	sprintf( dummyBuffer, "VGNDS2I %s 0 DC %g\n", gndPin_P->spiceNodeName, 
			 Gnd );
strcat( powerBuffer, dummyBuffer );

sprintf( temperatureBuffer, temperatureString[spiceType], temperature );

}


void
setupDCSweepCommands(	OUT	char	*analysisBuffer,
						IN	int		spiceType,
						IN	double	sweepStart,
						IN	double	sweepEnd,
						IN	double	sweepStep )

{
/*==== Declarations =====================================================*/
char	dummyBuffer[128];			/* temporary buffer */

/*==== Code =============================================================*/
sprintf( analysisBuffer, ".DC VOUTS2I %g %g %g\n", sweepStart, sweepEnd,
		 sweepStep );
sprintf( dummyBuffer, spicePrintVIString[spiceType] );
strcat( analysisBuffer, dummyBuffer );

}



void
setupTranCommands(	OUT	char	*analysisBuffer,
					IN	int		spiceType,
					IN	double	simTime,
					IN	char	*outputNode )

{
/*==== Declarations =====================================================*/
char	dummyBuffer[128];			/* temporary buffer */
char	nodeNameBuffer[128];	
int		dummy;

/*==== Code =============================================================*/
sprintf( analysisBuffer, ".TRAN %g %g \n", simTime / 100, simTime );

if ( ( spiceType == SPICE_TYPE_PSPICE ) && 
	 !( IS_AN_INTEGER_NODE_NAME( outputNode, dummy ) ) )
	sprintf( nodeNameBuffer, "[%s]", outputNode );
else
	sprintf( nodeNameBuffer, "%s", outputNode );
sprintf( dummyBuffer, spicePrintTranString[spiceType], nodeNameBuffer );
strcat( analysisBuffer, dummyBuffer );
}


void
setupSpiceFileNames(	OUT	char	*spiceInFile,
						OUT	char	*spiceOutFile,
						OUT	char	*spiceMsgFile,
						IN	char	*fileNamePrefix,
						IN	char	*pinName )

{
sprintf( spiceInFile, "%s%s.spi", fileNamePrefix, pinName );
sprintf( spiceOutFile, "%s%s.out", fileNamePrefix, pinName );
sprintf( spiceMsgFile, "%s%s.msg", fileNamePrefix, pinName );
}



int
setupSpiceInputFile(	IN	int		iterate,
						IN	int		spiceType,
						IN	char	*spiceInputFile,
						IN	char	*headerLine,
						IN	char	*spiceFile,
						IN	char	*modelFile,
						IN	char	*inputBuffer,
						IN	char	*powerBuffer,
						IN	char	*temperatureBuffer,
						IN	char	*analysisBuffer )

{
/*==== Declarations =====================================================*/
int		nextchar;					/* holds value from fgetc */
char	headerLine2[256],
		lineBuffer[256];
FILE	*spiceInputFile_P = NULL,	/* points to the Spice input file */
		*tempFile_P = NULL;			/* points to whichever file we need */

/*==== Code =============================================================*/
sprintf( headerLine2, "*\n* Spice deck created by s2ibis v %s\n* North Carolina State University\n*\n",
		 VERSION_STRING );

/*-------------------------------------------------------------------------
*
*	Only set up the file if iterate is not set or if the file does not
*	already exist.
*
*------------------------------------------------------------------------*/
if ( !( iterate ) || 
	 ( ( spiceInputFile_P = fopen( spiceInputFile, "r" ) ) == NULL ) )
	{
	/*---------------------------------------------------------------------
	*
	*	Now check to make sure we can open the file for writing.  If not,
	*	return an error.  If so, begin creating the spice file.
	*
	*--------------------------------------------------------------------*/
	if ( spiceInputFile_P != NULL )
		fclose( spiceInputFile_P );
	if ( ( spiceInputFile_P = fopen( spiceInputFile, "w" ) ) == NULL )
		{
		fprintf( stderr, "s2ibis2: Unable to open " );
		fprintf( stderr, "file %s for writing.\n", spiceInputFile );
		perror( "s2ibis2" );
		return( 1 );
		}
	else
		{
		/*-----------------------------------------------------------------
		*
		*	First, print out the header information.
		*
		*----------------------------------------------------------------*/
		if ( fprintf( spiceInputFile_P, "%s%s", headerLine, headerLine2 ) < 0 )
			{
			fprintf( stderr, "s2ibis2: Error while writing header lines " );
			fprintf( stderr, "in file %s.\n", spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}

		/*-----------------------------------------------------------------
		*
		*	Next, copy the component spice file to the spice input file.
		*
		*----------------------------------------------------------------*/
		if ( ( tempFile_P = fopen( spiceFile, "r" ) ) == NULL )
			{
			fprintf( stderr, "s2ibis2: Unable " );
			fprintf( stderr, "to open file %s for reading.\n", spiceFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		while ( fgets( lineBuffer, 256, tempFile_P ) != NULL )
		/*-----------------------------------------------------------------
		*	We need to check two conditions:  if the first column of the line
		*	is a *, we don't write the line because it's a comment; and if
		*	the line begins with ".END" and the fifth column is not a 'S', 
		*	we don't write it because it's the end of the file.  We also
		*	need to check the lowercase version.  Ugh.
		*----------------------------------------------------------------*/
			if ( ( lineBuffer[0] != '*' ) && 
				 ( ( strstr( lineBuffer, ".END" ) == NULL ) || 
				   ( lineBuffer[4] == 'S' ) ) && 
				 ( ( strstr( lineBuffer, ".end" ) == NULL ) || 
				   ( lineBuffer[4] == 's' ) ) )
				if ( fputs( lineBuffer, spiceInputFile_P ) == EOF )
					{
					fprintf( stderr, "s2ibis2: Error copying file %s ", 
							 spiceFile );
					fprintf( stderr, "to file %s.\n", spiceInputFile );
					perror( "s2ibis2" );
					return( 1 );
					}
		if ( fclose( tempFile_P ) == EOF )
			{
			fprintf( stderr, "s2ibis2: Unable to close file %s.\n", spiceFile );
			perror( "s2ibis2" );
			return( 1 );
			}

		/*-----------------------------------------------------------------
		*
		*	Next, copy the model file to the spice input file.  However,
		*	we only want to do this if a model file has been specified.
		*
		*----------------------------------------------------------------*/
		if ( strcmp( modelFile, "" ) && strcmp( modelFile, "NA" ) )
			{
			if ( ( tempFile_P = fopen( modelFile, "r" ) ) == NULL )
				{
				fprintf( stderr, "s2ibis2: Unable " );
				fprintf( stderr, "to open file %s for reading.\n", spiceFile );
				perror( "s2ibis2" );
				return( 1 );
				}
			while ( ( nextchar = fgetc( tempFile_P ) ) != EOF )
				if ( fputc( nextchar, spiceInputFile_P ) == EOF )
					{
					fprintf( stderr, "s2ibis2: Error copying file %s ", 
							 spiceFile );
					fprintf( stderr, "to file %s.\n", spiceInputFile );
					perror( "s2ibis2" );
					return( 1 );
					}
			if ( fclose( tempFile_P ) == EOF )
				{
				fprintf( stderr, "s2ibis2: Unable to close file %s.\n", 
						 spiceFile );
				perror( "s2ibis2" );
				return( 1 );
				}
			}

		/*-----------------------------------------------------------------
		*
		*	Next, copy the the power source command buffer to the spice 
		*	input file.
		*
		*----------------------------------------------------------------*/
		if ( fputs( powerBuffer, spiceInputFile_P ) == EOF )
			{
			fprintf( stderr, "s2ibis2: Error adding power sources " );
			fprintf( stderr, "to file %s.\n", spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		/*-----------------------------------------------------------------
		*
		*	Next, copy the the input stimulus command buffer to the spice 
		*	input file.
		*
		*----------------------------------------------------------------*/
		if ( fputs( inputBuffer, spiceInputFile_P ) == EOF )
			{
			fprintf( stderr, "s2ibis2: Error adding input stimulus " );
			fprintf( stderr, "to file %s.\n", spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		/*-----------------------------------------------------------------
		*
		*	Next, copy the the temperature command buffer and the options
		*	string to the spice input file.
		*
		*----------------------------------------------------------------*/
		if ( ( fputs( temperatureBuffer, spiceInputFile_P ) == EOF ) ||
			 ( fputs( spiceOptions[spiceType], spiceInputFile_P ) == EOF ) )
			{
			fprintf( stderr, "s2ibis2: Error adding Spice options " );
			fprintf( stderr, "to file %s.\n", spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		/*-----------------------------------------------------------------
		*
		*	Next, copy the the analysis command buffer and the VI print
		*	string to the spice input file.
		*
		*----------------------------------------------------------------*/
		if ( fputs( analysisBuffer, spiceInputFile_P ) == EOF )
			{
			fprintf( stderr, "s2ibis2: Error adding Spice analysis commands " );
			fprintf( stderr, "to file %s.\n", spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		/*-----------------------------------------------------------------
		*
		*	Finally, write the Spice .END command to the file and close it.
		*
		*----------------------------------------------------------------*/
		if ( fprintf( spiceInputFile_P, ".END\n" ) < 0 ) 
			{
			fprintf( stderr, "s2ibis2: Error completing file %s.\n", 
					 spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		if ( fclose( spiceInputFile_P ) == EOF )
			{
			fprintf( stderr, "s2ibis2: Unable to close file %s.\n", 
					 spiceInputFile );
			perror( "s2ibis2" );
			return( 1 );
			}
		}
	}
else
	fclose( spiceInputFile_P );

return( 0 );
}
	

void
setPinTran(	OUT	char		*pinBuffer,
			IN	ibisPin_Pt	pin_P,
			IN	int			pinType,
			IN	int			outputRising,
			IN	double		simTime,
			IN	char		*nodeLabel )

{
double	Vinl = 0, 
		Vinh = 0,
		risetime = 0,
		falltime = 0,
		pulsewidth = 0,
		period = 0;

if ( IS_USE_NA( pin_P->model_P->Vinl ) )
	if ( IS_ECL( pin_P->model_P->modelType ) )
		Vinl = VINL_ECL_DEFAULT;
	else
		Vinl = VINL_STD_DEFAULT;
else
	Vinl = pin_P->model_P->Vinl;
	
if ( IS_USE_NA( pin_P->model_P->Vinh ) )
	if ( IS_ECL( pin_P->model_P->modelType ) )
		Vinh = VINH_ECL_DEFAULT;
	else
		Vinh = VINH_STD_DEFAULT;
else
	Vinh = pin_P->model_P->Vinh;

risetime = simTime / 100;
falltime = simTime / 100;
pulsewidth = 2 * simTime;
period = 2 * simTime;
	
if ( ( pinType == MODEL_POLARITY_INVERTING ) ||
	 ( pinType == MODEL_ENABLE_ACTIVE_LOW ) )
	if ( outputRising )
		sprintf( pinBuffer, "V%sS2I %s 0 PULSE(%g %g 0 %g %g %g %g)\n", 
				 nodeLabel, pin_P->spiceNodeName, Vinh, Vinl, risetime,
				 falltime, pulsewidth, period );
	else
		sprintf( pinBuffer, "V%sS2I %s 0 PULSE(%g %g 0 %g %g %g %g)\n", 
				 nodeLabel, pin_P->spiceNodeName, Vinl, Vinh, risetime,
				 falltime, pulsewidth, period );
else
	if ( outputRising )
		sprintf( pinBuffer, "V%sS2I %s 0 PULSE(%g %g 0 %g %g %g %g)\n", 
				 nodeLabel, pin_P->spiceNodeName, Vinl, Vinh, risetime,
				 falltime, pulsewidth, period );
	else
		sprintf( pinBuffer, "V%sS2I %s 0 PULSE(%g %g 0 %g %g %g %g)\n", 
				 nodeLabel, pin_P->spiceNodeName, Vinh, Vinl, risetime,
				 falltime, pulsewidth, period );
}



void
setPinDC(	OUT	char		*pinBuffer,
			IN	ibisPin_Pt	pin_P,
			IN	int			pinType,
			IN	int			outputActive,
			IN	char		*nodeLabel )

{
double	Vinl, Vinh;

if ( IS_USE_NA( pin_P->model_P->Vinl ) )
	if ( IS_ECL( pin_P->model_P->modelType ) )
		Vinl = VINL_ECL_DEFAULT;
	else
		Vinl = VINL_STD_DEFAULT;
else
	Vinl = pin_P->model_P->Vinl;
	
if ( IS_USE_NA( pin_P->model_P->Vinh ) )
	if ( IS_ECL( pin_P->model_P->modelType ) )
		Vinh = VINH_ECL_DEFAULT;
	else
		Vinh = VINH_STD_DEFAULT;
else
	Vinh = pin_P->model_P->Vinh;
	
if ( ( pinType == MODEL_POLARITY_INVERTING ) ||
	 ( pinType == MODEL_ENABLE_ACTIVE_LOW ) )
	if ( outputActive )
		sprintf( pinBuffer, "V%sS2I %s 0 DC %g\n", nodeLabel,
				 pin_P->spiceNodeName, Vinl );
	else
		sprintf( pinBuffer, "V%sS2I %s 0 DC %g\n", nodeLabel,
				 pin_P->spiceNodeName, Vinh );
else
	if ( outputActive )
		sprintf( pinBuffer, "V%sS2I %s 0 DC %g\n", nodeLabel,
				 pin_P->spiceNodeName, Vinh );
	else
		sprintf( pinBuffer, "V%sS2I %s 0 DC %g\n", nodeLabel,
				 pin_P->spiceNodeName, Vinl );

}



int
callSpice(	IN	int		iterate,
			IN	char	*spiceCommand,
			IN	int		spiceType,
			IN	char	*spiceInFile,
			IN	char	*spiceOutFile,
			IN	char	*spiceMsgFile )

{
FILE	*tempFile_P;
char	commandBuffer[256];
int		status = 0;

if ( strcmp( spiceCommand, "" ) )
	sprintf( commandBuffer, spiceCommand, spiceInFile, spiceOutFile,
			 spiceMsgFile );
else
	sprintf( commandBuffer, defaultSpiceCommand[spiceType], spiceInFile,
			 spiceOutFile, spiceMsgFile );

if ( !( iterate ) || ( ( tempFile_P = fopen( spiceOutFile, "r" ) ) == NULL ) )
	{
	fprintf( stderr, "s2ibis2: Starting %s job with input %s.\n",
			 spiceNameString[spiceType], spiceInFile );
	status = system( commandBuffer );

	if ( status == 127 )
		{
		fprintf( stderr, "s2ibis2: Unable to execute Spice.\n" );
		perror( "s2ibis2" );
		fprintf( stderr, "s2ibis2: Fatal error.\n" );
		exit( 1 );
		}
	}
else
	{
	fprintf( stderr, "s2ibis2: [Iterate] set and file %s exists. ",
			 spiceOutFile );
	fprintf( stderr, "Skipping run.\n" );
	fclose( tempFile_P );
	}

return( 0 );
}



int
checkForAbort(	IN	int		spiceType,
				IN	char	*spiceOutFile,
				IN	char	*spiceMsgFile )

{
FILE	*tempFile_P = NULL;
char	lineBuffer[256];

if ( ( tempFile_P = fopen( spiceOutFile, "r" ) ) != NULL )
	{
	while ( fgets( lineBuffer, 256, tempFile_P ) != NULL )
		if ( strstr( lineBuffer, abortMarker1[spiceType] ) != NULL )
			return( 1 );
	if ( strcmp( abortMarker2[spiceType], "" ) )
		{
		rewind( tempFile_P );
		while ( fgets( lineBuffer, 256, tempFile_P ) != NULL )
			if ( strstr( lineBuffer, abortMarker2[spiceType] ) != NULL )
				return( 1 );
		}
	fclose( tempFile_P );
	}

if ( ( tempFile_P = fopen( spiceMsgFile, "r" ) ) != NULL )
	{
	while ( fgets( lineBuffer, 256, tempFile_P ) != NULL )
		if ( strstr( lineBuffer, abortMarker1[spiceType] ) != NULL )
			return( 1 );
	if ( strcmp( abortMarker2[spiceType], "" ) )
		{
		rewind( tempFile_P );
		while ( fgets( lineBuffer, 256, tempFile_P ) != NULL )
			if ( strstr( lineBuffer, abortMarker2[spiceType] ) != NULL )
				return( 1 );
		}
	fclose( tempFile_P );
	}

return( 0 );
}



int
getSpiceVIData(	IN	ibisVItable_Pt	VIData_P,
				IN	int				tableSize,
				IN	int				spiceType,
				IN	char			*spiceOutFile,
				IN	char			*command )


{
FILE	*outFile_P = NULL;
char	lineBuffer[256];
double	tempData[300][2];
int		rowIndex = 1,
		i = 0;

if ( ( outFile_P = fopen( spiceOutFile, "r" ) ) == NULL )
	{
	fprintf( stderr, "s2ibis2: Unable to open file %s for reading.\n",
			 spiceOutFile );
	perror( "s2ibis2");
	return( 1 );
	}

/*------------------------------------------------------------------------
*
*	Read lines until we hit the data begin marker or EOF.  If it's EOF, 
*	return an error.
*
*-----------------------------------------------------------------------*/
while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( strstr( lineBuffer, VIDataBeginMarker[spiceType] ) == NULL ) );

if ( strstr( lineBuffer, VIDataBeginMarker[spiceType] ) == NULL )
	{
	fprintf( stderr, "s2ibis2: Data begin marker %s not found ",
			 VIDataBeginMarker[spiceType] );
	fprintf( stderr, "in output file %s.\n", spiceOutFile );
	return( 1 );
	}

/*------------------------------------------------------------------------
*
*	We're now at the line before the start of the VI data.  
*	Read lines until we get the correct number of columns.
*
*-----------------------------------------------------------------------*/
while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( sscanf( lineBuffer, VIDataConversionString[spiceType], 
				  &tempData[0][0], &tempData[0][1] ) != 2 ) );

if ( feof( outFile_P ) )
	{
	fprintf( stderr, "s2ibis2: No data in file %s\n", spiceOutFile );
	return( 1 );
	}

tempData[0][1] = -tempData[0][1];

/*------------------------------------------------------------------------
*
*	Now we read until we see the end-of-data marker or the end-of-file 
*	marker.  We include the data from any read that returns the correct 
*	number of columns read.  If we don't have an end-of-data marker for
*	this particular flavor of Spice, we just read to the end-of-file maker.
*
*-----------------------------------------------------------------------*/
if ( dataEndMarker[spiceType] == NULL )
	{
	while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) )
		if ( sscanf(lineBuffer, VIDataConversionString[spiceType],
		   &tempData[rowIndex][0], &tempData[rowIndex][1]) == 2 )
			{
			tempData[rowIndex][1] = -tempData[rowIndex][1];
			rowIndex++;
			}
	}
else
	{
	while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
			( strstr( lineBuffer, dataEndMarker[spiceType] ) == NULL ) ) 
		if ( sscanf(lineBuffer, VIDataConversionString[spiceType],
		   &tempData[rowIndex][0], &tempData[rowIndex][1]) == 2 )
			{
			tempData[rowIndex][1] = -tempData[rowIndex][1];
			rowIndex++;
			}
	}

fclose( outFile_P );

rowIndex--;
if ( rowIndex > tableSize )
	{
	fprintf( stderr, "s2ibis2: Spice job returned too many data points.\n" );
	fprintf( stderr, "         May have to truncate table...\n" );
	}

/*------------------------------------------------------------------------
*
*	Now we transfer the data to the correct place, depending on the value
*	of command.
*
*-----------------------------------------------------------------------*/
if ( !( strcmp( command, "typ" ) ) )
	{
	VIData_P->size = rowIndex + 1;
	for ( i = 0; i <= rowIndex; i++ )
		{
		VIData_P->VIs[i].v = tempData[i][0];
		VIData_P->VIs[i].i.typ = tempData[i][1];
		}
	}
else if ( !( strcmp( command, "min" ) ) )
	for ( i = 0; i <= rowIndex; i++ )
		VIData_P->VIs[i].i.min = tempData[i][1];
else
	for ( i = 0; i <= rowIndex; i++ )
		VIData_P->VIs[i].i.max = tempData[i][1];

return( 0 );
}


int
generateVICurve(	OUT	ibisVItable_Pt	*VIData_PP,
					IN	ibisPin_Pt		currentPin_P,
					IN	ibisPin_Pt		enablePin_P,
					IN	ibisPin_Pt		inputPin_P,
					IN	ibisPin_Pt		powerPin_P,
					IN	ibisPin_Pt		gndPin_P,
					IN	ibisTypMinMax_t	Vcc,
					IN	ibisTypMinMax_t	Gnd,
					IN	ibisTypMinMax_t	sweepStart,
					IN	double			sweepRange,
					IN	double			sweepStep,
					IN	int				curveType,
					IN	int				spiceType,
					IN	char			*spiceFile,
					IN	char			*spiceCommand,
					IN	int				enableOutput,
					IN	int				outputHigh,
					IN	int				iterate,
					IN	int				cleanup )

{
ibisModel_Pt	model_P;
double			sweepEnd = 0;
int				numTablePoints = 0;
char			headerLine[128],
				inputBuffer[256],
				dummyBuffer[256],
				powerBuffer[256],
				temperatureBuffer[128],
				analysisBuffer[256],
				spiceInFile[MAX_FILENAME_LENGTH + 1],
				spiceOutFile[MAX_FILENAME_LENGTH + 1],
				spiceMsgFile[MAX_FILENAME_LENGTH + 1];

				


model_P = currentPin_P->model_P; 

numTablePoints = (int)( sweepRange / sweepStep ) + 1;

if ( create_VItable( VIData_PP, numTablePoints ) )
	return( 1 );

if ( ( enablePin_P != NULL ) && ( enableOutput ) )
	setPinDC( inputBuffer, enablePin_P, model_P->enable, enableOutput, "ENA" );
else
	sprintf( inputBuffer, "" );

if ( inputPin_P != NULL )
	{
	setPinDC( dummyBuffer, inputPin_P, model_P->polarity, outputHigh, "IN" );
	strcat( inputBuffer, dummyBuffer );
	}

sprintf( dummyBuffer, "VOUTS2I %s 0 DC 0\n", currentPin_P->spiceNodeName );
strcat( inputBuffer, dummyBuffer );

/*------------------------------------------------------------------------
*
*	Set up for the typ run, and do the run.
*
*-----------------------------------------------------------------------*/
sweepEnd = sweepStart.typ + sweepRange;

sprintf( headerLine, "* Typ %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.typ, Gnd.typ, 
						   model_P->tempRange.typ );

setupDCSweepCommands( analysisBuffer, spiceType, sweepStart.typ, sweepEnd, 
					  sweepStep);

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 spiceFileTypPrefix[curveType], currentPin_P->pinName );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileTyp, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (TYP) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	fprintf( stderr, "         Curve %s not generated.\n", 
			 curveNameString[curveType] );
	delete_VItable( VIData_PP );
	return( 1 );
	}
else
	getSpiceVIData(*VIData_PP, numTablePoints, spiceType, spiceOutFile, "typ"); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}

/*------------------------------------------------------------------------
*
*	Set up for the min run, and do the run.
*
*-----------------------------------------------------------------------*/
sweepEnd = sweepStart.min + sweepRange;

sprintf( headerLine, "* Min %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.min, Gnd.min, 
						   model_P->tempRange.min );

setupDCSweepCommands( analysisBuffer, spiceType, sweepStart.min, sweepEnd, 
					  sweepStep);

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 spiceFileMinPrefix[curveType], currentPin_P->pinName );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileMin, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (MIN) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	fprintf( stderr, "         continuing with MAX run.\n");
	}
else
	getSpiceVIData(*VIData_PP, numTablePoints, spiceType, spiceOutFile, "min" );

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}

/*------------------------------------------------------------------------
*
*	Set up for the max run, and do the run.
*
*-----------------------------------------------------------------------*/
sweepEnd = sweepStart.max + sweepRange;

sprintf( headerLine, "* Max %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.max, Gnd.max, 
						   model_P->tempRange.max );

setupDCSweepCommands( analysisBuffer, spiceType, sweepStart.max, sweepEnd, 
					  sweepStep);

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile, 
					 spiceFileMaxPrefix[curveType], currentPin_P->pinName );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileMax, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (MAX) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	}
else
	getSpiceVIData(*VIData_PP, numTablePoints, spiceType, spiceOutFile, "max" );

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}

return( 0 );
}



int
getSpiceRampData(	INOUT	ibisModel_Pt	model_P,
					IN		int				curveType,
					IN		int				spiceType,
					IN		char			*spiceOutFile,
					IN		char			*command )


{
FILE	*outFile_P = NULL;
char	lineBuffer[256];
double	t0 = 0,
		t1 = 0,
		t20 = 0,
		t80 = 0,
		v0 = 0,
		v1 = 0,
		v20 = 0,
		v80 = 0;

if ( ( outFile_P = fopen( spiceOutFile, "r" ) ) == NULL )
	{
	fprintf( stderr, "s2ibis2: Unable to open file %s for reading.\n",
			 spiceOutFile );
	perror( "s2ibis2");
	return( 1 );
	}

/*------------------------------------------------------------------------
*	Read lines until we hit the data begin marker or EOF.  If it's EOF, 
*	return an error.
*-----------------------------------------------------------------------*/
while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( strstr( lineBuffer, tranDataBeginMarker[spiceType] ) == NULL ) );

if ( strstr( lineBuffer, tranDataBeginMarker[spiceType] ) == NULL )
	{
	fprintf( stderr, "s2ibis2: Data begin marker %s not found ",
			 tranDataBeginMarker[spiceType] );
	fprintf( stderr, "in output file %s.\n", spiceOutFile );
	return( 1 );
	}

/*------------------------------------------------------------------------
*	We're now at the line before the start of the transient data.  Read until
*	you hit the first line of data.
*-----------------------------------------------------------------------*/
while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( sscanf( lineBuffer, tranDataConversionString[spiceType], 
				  &t0, &v0 ) != 2 ) );

if ( feof( outFile_P ) )
	{
	fprintf( stderr, "s2ibis2: No data in file %s\n", spiceOutFile );
	return( 1 );
	}

/*------------------------------------------------------------------------
*	Now we read until we no longer get the correct number of columns.  We
*	use a method similar to that found in getSpiceVIData.
*-----------------------------------------------------------------------*/
if ( dataEndMarker[spiceType] == NULL )
	{
	while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) )
		sscanf( lineBuffer, tranDataConversionString[spiceType], &t1, &v1 );
	}
else
	{
	while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
			( strstr( lineBuffer, dataEndMarker[spiceType] ) == NULL ) ) 
		sscanf( lineBuffer, tranDataConversionString[spiceType], &t1, &v1 );
	}

/*------------------------------------------------------------------------
*	Determine the 20% and 80% voltage points.
*-----------------------------------------------------------------------*/
v20 = v0 + 0.2 * ( v1 - v0 );
v80 = v0 + 0.8 * ( v1 - v0 );

/*------------------------------------------------------------------------
*	Now we start at the beginning of the file, and look for the 20% voltage.
*	When we find it, we mark that time as the 20% time.  Same thing for the
*	80% voltage.  Do this with linear interpolation, to get at least
*	reasonable accuracy for the times.
*-----------------------------------------------------------------------*/
rewind( outFile_P );

while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( strstr( lineBuffer, tranDataBeginMarker[spiceType] ) == NULL ) );

while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( sscanf( lineBuffer, tranDataConversionString[spiceType], 
				  &t0, &v0 ) != 2 ) );

if ( v1 > v0 )
	{
	v1 = v0;
	if ( v0 >= v20 )
		t20 = t0;
	else
		{
		while ( ( v1 < v20 ) && 
				( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
				( sscanf( lineBuffer, tranDataConversionString[spiceType], 
							&t1, &v1 ) == 2 ) )
			if ( v1 >= v20 )
				t20 = t0 + ( ( v20 - v0 ) * ( t1 - t0 ) / ( v1 - v0 ) );
			else
				{
				v0 = v1;
				t0 = t1;
				}
		while ( ( v1 < v80 ) &&
				( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
				( sscanf( lineBuffer, tranDataConversionString[spiceType], 
							&t1, &v1 ) == 2 ) )
			if ( v1 >= v80 )
				t80 = t0 + ( ( v80 - v0 ) * ( t1 - t0 ) / ( v1 - v0 ) );
			else
				{
				v0 = v1;
				t0 = t1;
				}
		}
	}
else
	{
	v1 = v0;
	if ( v0 <= v20 )
		t20 = t0;
	else
		{
		while ( ( v1 > v20 ) && 
				( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
				( sscanf( lineBuffer, tranDataConversionString[spiceType], 
							&t1, &v1 ) == 2 ) )
			if ( v1 <= v20 )
				t20 = t1;
		while ( ( v1 > v80 ) &&
				( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
				( sscanf( lineBuffer, tranDataConversionString[spiceType], 
							&t1, &v1 ) == 2 ) )
			if ( v1 <= v80 )
				t80 = t1;
		}
	}

/*------------------------------------------------------------------------
*	Now we figure out where we want the data to go, munge it up, and 
*	store it.
*-----------------------------------------------------------------------*/
if ( curveType == CURVE_TYPE_RISING_RAMP )
	{
	if ( !( strcmp( command, "typ" ) ) )
		{
		model_P->ramp.dv_r.typ = v80 - v20;
		model_P->ramp.dt_r.typ = t80 - t20;
		}
	else if ( !( strcmp( command, "min" ) ) )
		{
		model_P->ramp.dv_r.min = v80 - v20;
		model_P->ramp.dt_r.min = t80 - t20;
		model_P->ramp.dt_r.min -= model_P->ramp.dt_r.min * 
								  ( model_P->derateRampPct / 100 );
		}
	else
		{
		model_P->ramp.dv_r.max = v80 - v20;
		model_P->ramp.dt_r.max = t80 - t20;
		model_P->ramp.dt_r.max += model_P->ramp.dt_r.max * 
								  ( model_P->derateRampPct / 100 );
		}
	}
else
	{
	if ( !( strcmp( command, "typ" ) ) )
		{
		model_P->ramp.dv_f.typ = v20 - v80;
		model_P->ramp.dt_f.typ = t80 - t20;
		}
	else if ( !( strcmp( command, "min" ) ) )
		{
		model_P->ramp.dv_f.min = v20 - v80;
		model_P->ramp.dt_f.min = t80 - t20;
		model_P->ramp.dt_f.min -= model_P->ramp.dt_f.min * 
								  ( model_P->derateRampPct / 100 );
		}
	else
		{
		model_P->ramp.dv_f.max = v20 - v80;
		model_P->ramp.dt_f.max = t80 - t20;
		model_P->ramp.dt_f.max += model_P->ramp.dt_f.max * 
								  ( model_P->derateRampPct / 100 );
		}
	}

return( 0 );
}



int
generateRampData(	INOUT	ibisPin_Pt		currentPin_P,
					IN		ibisPin_Pt		enablePin_P,
					IN		ibisPin_Pt		inputPin_P,
					IN		ibisPin_Pt		powerPin_P,
					IN		ibisPin_Pt		gndPin_P,
					IN		ibisTypMinMax_t	Vcc,
					IN		ibisTypMinMax_t	Gnd,
					IN		int				curveType,
					IN		int				spiceType,
					IN		char			*spiceFile,
					IN		char			*spiceCommand,
					IN		int				iterate,
					IN		int				cleanup )

{

ibisModel_Pt	model_P;
char			headerLine[128],
				inputBuffer[512],
				dummyBuffer[256],
				powerBuffer[256],
				temperatureBuffer[128],
				analysisBuffer[256],
				spiceInFile[MAX_FILENAME_LENGTH + 1],
				spiceOutFile[MAX_FILENAME_LENGTH + 1],
				spiceMsgFile[MAX_FILENAME_LENGTH + 1];
int				outputState = 0;


model_P = currentPin_P->model_P; 

if ( curveType == CURVE_TYPE_RISING_RAMP )
	outputState = OUTPUT_RISING;
else
	outputState = OUTPUT_FALLING;

/*------------------------------------------------------------------------
*	First enable the output.
*-----------------------------------------------------------------------*/
if ( enablePin_P != NULL )
	setPinDC( inputBuffer, enablePin_P, model_P->enable, ENABLE_OUTPUT, "ENA" );
else
	sprintf( inputBuffer, "" );

/*------------------------------------------------------------------------
*	Now, we attach the load to the output and the proper termination 
*	voltage.
*-----------------------------------------------------------------------*/
if ( ( model_P->modelType == MODEL_TYPE_OPEN_DRAIN ) ||
	 ( model_P->modelType == MODEL_TYPE_OPEN_SINK ) ||
	 ( model_P->modelType == MODEL_TYPE_I_O_OPEN_DRAIN ) ||
	 ( model_P->modelType == MODEL_TYPE_I_O_OPEN_SINK ) )
	{
	sprintf( dummyBuffer, "RLOADS2I %s %s %6.2g\n", currentPin_P->spiceNodeName,
			 powerPin_P->spiceNodeName, model_P->Rload );
	strcat( inputBuffer, dummyBuffer );
	}
else if ( ( model_P->modelType == MODEL_TYPE_OPEN_SOURCE ) ||
	 	  ( model_P->modelType == MODEL_TYPE_I_O_OPEN_SOURCE ) )
	{
	sprintf( dummyBuffer, "RLOADS2I %s %s %6.2g\n", currentPin_P->spiceNodeName,
			 gndPin_P->spiceNodeName, model_P->Rload );
	strcat( inputBuffer, dummyBuffer );
	}
else if ( ( model_P->modelType == MODEL_TYPE_OUTPUT_ECL ) ||
	 	  ( model_P->modelType == MODEL_TYPE_I_O_ECL ) )
	{
	sprintf( dummyBuffer, "RLOADS2I %s %s %6.2g\n", currentPin_P->spiceNodeName,
			 dummyNodeName[0], model_P->Rload );
	strcat( inputBuffer, dummyBuffer );
	sprintf( dummyBuffer, "VTERMS2I %s %s %6.2g\n", dummyNodeName[0],
			 powerPin_P->spiceNodeName, ECL_TERMINATION_VOLTAGE_DEFAULT );
	strcat( inputBuffer, dummyBuffer );
	}
else if ( curveType == CURVE_TYPE_RISING_RAMP )
	{
	sprintf( dummyBuffer, "RLOADS2I %s %s %6.2g\n", currentPin_P->spiceNodeName,
			 gndPin_P->spiceNodeName, model_P->Rload );
	strcat( inputBuffer, dummyBuffer );
	}
else
	{
	sprintf( dummyBuffer, "RLOADS2I %s %s %6.2g\n", currentPin_P->spiceNodeName,
			 powerPin_P->spiceNodeName, model_P->Rload );
	strcat( inputBuffer, dummyBuffer );
	}


/*------------------------------------------------------------------------
*	Now attach the stimulus to the input pin.
*-----------------------------------------------------------------------*/
if ( inputPin_P != NULL )
	{
	setPinTran( dummyBuffer, inputPin_P, model_P->polarity, outputState, 
				model_P->simTime, "IN" );
	strcat( inputBuffer, dummyBuffer );
	}

/*------------------------------------------------------------------------
*	Set up for the typ run, and do the run.
*-----------------------------------------------------------------------*/
sprintf( headerLine, "* Typ %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.typ, Gnd.typ, 
						   model_P->tempRange.typ );

setupTranCommands( analysisBuffer, spiceType, model_P->simTime, 
				   currentPin_P->spiceNodeName ); 

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 spiceFileTypPrefix[curveType], currentPin_P->pinName );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileTyp, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (TYP) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	fprintf( stderr, "         Curve %s not generated.\n", 
			 curveNameString[curveType] );
	return( 1 );
	}
else
	getSpiceRampData( model_P, curveType, spiceType, spiceOutFile, "typ" ); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}
	
/*------------------------------------------------------------------------
*	Set up for the min run, and do the run.
*-----------------------------------------------------------------------*/
sprintf( headerLine, "* Min %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.min, Gnd.min, 
						   model_P->tempRange.min );

setupTranCommands( analysisBuffer, spiceType, model_P->simTime, 
				   currentPin_P->spiceNodeName ); 

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 spiceFileMinPrefix[curveType], currentPin_P->pinName );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileMin, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (MIN) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	}
else
	getSpiceRampData( model_P, curveType, spiceType, spiceOutFile, "min" ); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}
						
/*------------------------------------------------------------------------
*	Set up for the max run, and do the run.
*-----------------------------------------------------------------------*/
sprintf( headerLine, "* Max %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.max, Gnd.max, 
						   model_P->tempRange.max );

setupTranCommands( analysisBuffer, spiceType, model_P->simTime, 
				   currentPin_P->spiceNodeName ); 

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 spiceFileMaxPrefix[curveType], currentPin_P->pinName );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileMax, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (MAX) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	}
else
	getSpiceRampData( model_P, curveType, spiceType, spiceOutFile, "max" ); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}

return( 0 );
}



void
binTranData( 	IN		double				t,
				IN		double				v,
				IN		double				simTime,
				IN		double				binTime,
				IN		char				*command,
				INOUT	int					*lastBin,
				INOUT	double				*runningSum,
				INOUT	int					*numPointsInBin,
				INOUT	ibisWaveTable_Pt	wave_P )

{
int		currentBin = 0,
		i = 0;
double	vLastBin = 0,
		tLastBin = 0,
		vInterp = 0,
		tInterp = 0;

currentBin = ROUND( t / binTime );
if ( currentBin == *lastBin )
	{
	/*--------------------------------------------------------
	*	If we're still in the last bin, just add this data 
	*	point to the running sum and continue.
	*-------------------------------------------------------*/
	( *runningSum ) += v;
	( *numPointsInBin ) += 1;
	}
else
	{
	/*--------------------------------------------------------
	*	If we've changed bins, we have some work to do.  First, 
	*	we compute the average of all the points in the bin,
	*	and assign this value to the appropriate spot in the
	*	wave table.
	*-------------------------------------------------------*/
	vLastBin = ( *runningSum ) / ( *numPointsInBin );
	tLastBin = ( *lastBin ) * binTime;
	if ( !( strcmp( command, "typ") ) )
		{
		wave_P->waveData[*lastBin].t = tLastBin;
		wave_P->waveData[*lastBin].v.typ = vLastBin;
		}
	else if ( !( strcmp( command, "min" ) ) )
		wave_P->waveData[*lastBin].v.min = vLastBin;
	else
		wave_P->waveData[*lastBin].v.max = vLastBin;

	/*--------------------------------------------------------
	*	Now we need to check if we've missed any bins.  If we 
	*	have, we're just going to linearly interpolate
	*	between the two adjacent points (for however many bins
	*	it takes) to get data points in the empty bins.
	*-------------------------------------------------------*/
	if ( currentBin > ( *lastBin ) + 1 )
		for ( i = ( *lastBin ) + 1; i < currentBin; i++ )
			{
			tInterp = i * binTime;
			vInterp = vLastBin + ( tInterp - tLastBin ) * 
					  ( v - vLastBin ) / ( t - tLastBin );
			if ( !( strcmp( command, "typ") ) )
				{
				wave_P->waveData[i].t = tInterp;
				wave_P->waveData[i].v.typ = vInterp;
				}
			else if ( !( strcmp( command, "min" ) ) )
				wave_P->waveData[i].v.min = vInterp;
			else
				wave_P->waveData[i].v.max = vInterp;
			}
			
	/*--------------------------------------------------------
	*	Finally, we reset our counters and running sum.
	*-------------------------------------------------------*/
	*lastBin = currentBin;
	*runningSum = v;
	*numPointsInBin = 1;
	}

/*------------------------------------------------------------
*	Put the last data point into the wave table.
*-----------------------------------------------------------*/
vLastBin = ( *runningSum ) / ( *numPointsInBin );
tLastBin = simTime;

if ( !( strcmp( command, "typ") ) )
	{
	wave_P->waveData[*lastBin].t = tLastBin;
	wave_P->waveData[*lastBin].v.typ = vLastBin;
	}
else if ( !( strcmp( command, "min" ) ) )
	wave_P->waveData[*lastBin].v.min = vLastBin;
else
	wave_P->waveData[*lastBin].v.max = vLastBin;
}



int
getSpiceWaveData(	INOUT	ibisWaveTable_Pt	wave_P,
					IN		double				simTime,
					IN		int					spiceType,
					IN		char				*spiceOutFile,
					IN		char				*command )


{
FILE		*outFile_P = NULL;
char		lineBuffer[256];
double		t = 0,
			v = 0,
			binTime = 0,
			runningSum = 0;
int			lastBin = 0,
			numPointsInBin = 0;

binTime = simTime / ( WAVE_POINTS_DEFAULT - 1 );

resizeWaveTable( wave_P, WAVE_POINTS_DEFAULT );

if ( ( outFile_P = fopen( spiceOutFile, "r" ) ) == NULL )
	{
	fprintf( stderr, "s2ibis2: Unable to open file %s for reading.\n",
			 spiceOutFile );
	perror( "s2ibis2");
	return( 1 );
	}

/*------------------------------------------------------------------------
*	Read lines until we hit the data begin marker or EOF.  If it's EOF, 
*	return an error.
*-----------------------------------------------------------------------*/
while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( strstr( lineBuffer, tranDataBeginMarker[spiceType] ) == NULL ) );

if ( strstr( lineBuffer, tranDataBeginMarker[spiceType] ) == NULL )
	{
	fprintf( stderr, "s2ibis2: Data begin marker %s not found ",
			 tranDataBeginMarker[spiceType] );
	fprintf( stderr, "in output file %s.\n", spiceOutFile );
	return( 1 );
	}

/*------------------------------------------------------------------------
*	We're now at the line before the start of the transient data.  Read until
*	you hit the first line of data.
*-----------------------------------------------------------------------*/
while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
		( sscanf( lineBuffer, tranDataConversionString[spiceType], 
				  &t, &v ) != 2 ) );

if ( feof( outFile_P ) )
	{
	fprintf( stderr, "s2ibis2: No data in file %s\n", spiceOutFile );
	return( 1 );
	}

numPointsInBin = 1;
runningSum = v;
lastBin = 0;

/*------------------------------------------------------------------------
*	Now we read until we no longer get the correct number of columns.
*	While we do this, we're going to bin the points to construct a curve
*	with evenly spaced time values.
*-----------------------------------------------------------------------*/
/*------------------------------------------------------------------------
*	Now we read until we no longer get the correct number of columns.  We
*	use a method similar to that found in getSpiceVIData.  While we do
*	this, we're going to bin the points to construct a curve with
*	evenly spaced time values.
*-----------------------------------------------------------------------*/
if ( dataEndMarker[spiceType] == NULL )
	{
	while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) )
		if ( sscanf( lineBuffer, tranDataConversionString[spiceType], 
					 &t, &v ) == 2 )
			{
			binTranData( t, v, simTime, binTime, command, &lastBin, 
						 &runningSum, &numPointsInBin, wave_P );
			}
	}
else
	{
	while ( ( fgets( lineBuffer, 256, outFile_P ) != NULL ) &&
			( strstr( lineBuffer, dataEndMarker[spiceType] ) == NULL ) ) 
		if ( sscanf( lineBuffer, tranDataConversionString[spiceType], 
					 &t, &v ) == 2 )
			{
			binTranData( t, v, simTime, binTime, command, &lastBin, 
						 &runningSum, &numPointsInBin, wave_P );
			}
	}

return( 0 );
}


int
generateWaveData(	INOUT	ibisWaveTable_Pt	currentWave_P,
					IN		ibisPin_Pt			currentPin_P,
					IN		ibisPin_Pt			enablePin_P,
					IN		ibisPin_Pt			inputPin_P,
					IN		ibisPin_Pt			powerPin_P,
					IN		ibisPin_Pt			gndPin_P,
					IN		ibisTypMinMax_t		Vcc,
					IN		ibisTypMinMax_t		Gnd,
					IN		int					curveType,
					IN		int					spiceType,
					IN		char				*spiceFile,
					IN		char				*spiceCommand,
					IN		int					iterate,
					IN		int					cleanup,
					IN		int					index )

{

ibisModel_Pt	model_P;
char			headerLine[128],
				inputBuffer[1024],
				inputBuffer2[1024],
				dummyBuffer[256],
				powerBuffer[256],
				temperatureBuffer[128],
				analysisBuffer[256],
				spiceInFile[MAX_FILENAME_LENGTH + 1],
				spiceOutFile[MAX_FILENAME_LENGTH + 1],
				spiceMsgFile[MAX_FILENAME_LENGTH + 1],
				outputNodeName[128];
char			*nodeList[11];
int				outputState = 0,
				nodeIndex = 0,
				i = 0;


model_P = currentPin_P->model_P; 

if ( curveType == CURVE_TYPE_RISING_WAVE )
	outputState = OUTPUT_RISING;
else
	outputState = OUTPUT_FALLING;

/*------------------------------------------------------------------------
*	First enable the output.
*-----------------------------------------------------------------------*/
if ( enablePin_P != NULL )
	setPinDC( inputBuffer, enablePin_P, model_P->enable, ENABLE_OUTPUT, "ENA" );
else
	sprintf( inputBuffer, "" );

/*------------------------------------------------------------------------
*	Now, we attach the load to the output.  This is both simpler and more
*	complex than for the ramp--there are no special cases for different
*	types of circuits, but there may be variable numbers of parasitics to
*	attach.  Ugh.  We start by building a list of node names--the first
*	node name in the list is the current pin's node, and the rest are the
*	dummy node names from s2istrng.h.
*-----------------------------------------------------------------------*/
nodeList[0] = currentPin_P->spiceNodeName;
for ( i = 1; i < 11; i++)
	nodeList[i] = dummyNodeName[i-1];

/*------------------------------------------------------------------------
*	Now we attach the __dut parasitics.  The measurement port comes between
*	the dut stuff and the fixture parasitics, so we'll save that node name.
*-----------------------------------------------------------------------*/
if ( !( IS_USE_NA( currentWave_P->L_dut ) ) )
	{
	sprintf( dummyBuffer, "LDUTS2I %s %s %6.2g\n", nodeList[nodeIndex],
			 nodeList[++nodeIndex], currentWave_P->L_dut );
	strcat( inputBuffer, dummyBuffer );
	}
if ( !( IS_USE_NA( currentWave_P->R_dut ) ) )
	{
	sprintf( dummyBuffer, "RDUTS2I %s %s %6.2g\n", nodeList[nodeIndex],
			 nodeList[++nodeIndex], currentWave_P->R_dut );
	strcat( inputBuffer, dummyBuffer );
	}
if ( !( IS_USE_NA( currentWave_P->C_dut ) ) )
	{
	sprintf( dummyBuffer, "CDUTS2I %s 0 %6.2g\n", nodeList[nodeIndex],
			 currentWave_P->C_dut );
	strcat( inputBuffer, dummyBuffer );
	}

strcpy( outputNodeName, nodeList[nodeIndex] );

/*------------------------------------------------------------------------
*	Now attach the fixture parasitics.  We'll attach v_fixture later.
*-----------------------------------------------------------------------*/
if ( !( IS_USE_NA( currentWave_P->L_fixture ) ) )
	{
	sprintf( dummyBuffer, "LFIXS2I %s %s %6.2g\n", nodeList[nodeIndex],
			 nodeList[++nodeIndex], currentWave_P->L_fixture );
	strcat( inputBuffer, dummyBuffer );
	}
if ( !( IS_USE_NA( currentWave_P->C_fixture ) ) )
	{
	sprintf( dummyBuffer, "CFIXS2I %s 0 %6.2g\n", nodeList[nodeIndex],
			 currentWave_P->C_fixture );
	strcat( inputBuffer, dummyBuffer );
	}

/*------------------------------------------------------------------------
*	R_fixture is always attached...
*-----------------------------------------------------------------------*/
sprintf( dummyBuffer, "RFIXS2I %s %s %6.2g\n", nodeList[nodeIndex],
		 nodeList[++nodeIndex], currentWave_P->R_fixture );
strcat( inputBuffer, dummyBuffer );

/*------------------------------------------------------------------------
*	Now attach the stimulus to the input pin.
*-----------------------------------------------------------------------*/
if ( inputPin_P != NULL )
	{
	setPinTran( dummyBuffer, inputPin_P, model_P->polarity, outputState, 
				model_P->simTime, "IN" );
	strcat( inputBuffer, dummyBuffer );
	}

strcpy( inputBuffer2, inputBuffer );

/*------------------------------------------------------------------------
*	Set up for the typ run, and do the run.  This is where we'll attach
*	V_fixture, since it may change with the min and max runs.
*-----------------------------------------------------------------------*/
sprintf( headerLine, "* Typ %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.typ, Gnd.typ, 
						   model_P->tempRange.typ );

setupTranCommands( analysisBuffer, spiceType, model_P->simTime, 
				   outputNodeName ); 
if ( index < 10 )
	sprintf( dummyBuffer, "%s0%d", spiceFileTypPrefix[curveType], index );
else
	sprintf( dummyBuffer, "%s%d", spiceFileTypPrefix[curveType], index );

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 dummyBuffer, currentPin_P->pinName );

sprintf( dummyBuffer, "VFIXS2I %s 0 %6.2g\n", nodeList[nodeIndex],
		 currentWave_P->V_fixture );
strcat( inputBuffer, dummyBuffer );

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileTyp, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (TYP) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	fprintf( stderr, "         Curve %s not generated.\n", 
			 curveNameString[curveType] );
	return( 1 );
	}
else
	getSpiceWaveData( currentWave_P, model_P->simTime, spiceType, 
					  spiceOutFile, "typ" ); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}
	
/*------------------------------------------------------------------------
*	Set up for the min run, and do the run.
*-----------------------------------------------------------------------*/
sprintf( headerLine, "* Min %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.min, Gnd.min, 
						   model_P->tempRange.min );

setupTranCommands( analysisBuffer, spiceType, model_P->simTime, 
				   outputNodeName ); 
if ( index < 10 )
	sprintf( dummyBuffer, "%s0%d", spiceFileMinPrefix[curveType], index );
else
	sprintf( dummyBuffer, "%s%d", spiceFileMinPrefix[curveType], index );

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 dummyBuffer, currentPin_P->pinName );

if ( !( IS_USE_NA( currentWave_P->V_fixture_min ) ) )
	{
	strcpy( inputBuffer, inputBuffer2 );
	sprintf( dummyBuffer, "VFIXS2I %s 0 %6.2g\n", nodeList[nodeIndex],
			 currentWave_P->V_fixture_min );
	strcat( inputBuffer, dummyBuffer );
	}

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileMin, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (MIN) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	}
else
	getSpiceWaveData( currentWave_P, model_P->simTime, spiceType, 
					  spiceOutFile, "min" ); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}
	
/*------------------------------------------------------------------------
*	Set up for the max run, and do the run.
*-----------------------------------------------------------------------*/
sprintf( headerLine, "* Max %s curve for model %s\n", 
		 curveNameString[curveType], model_P->modelName );

setupPowerAndTempCommands( powerBuffer, temperatureBuffer, spiceType,
						   powerPin_P, gndPin_P, Vcc.max, Gnd.max, 
						   model_P->tempRange.max );

setupTranCommands( analysisBuffer, spiceType, model_P->simTime, 
				   outputNodeName ); 
if ( index < 10 )
	sprintf( dummyBuffer, "%s0%d", spiceFileMaxPrefix[curveType], index );
else
	sprintf( dummyBuffer, "%s%d", spiceFileMaxPrefix[curveType], index );

setupSpiceFileNames( spiceInFile, spiceOutFile, spiceMsgFile,
					 dummyBuffer, currentPin_P->pinName );

if ( !( IS_USE_NA( currentWave_P->V_fixture_max ) ) )
	{
	strcpy( inputBuffer, inputBuffer2 );
	sprintf( dummyBuffer, "VFIXS2I %s 0 %6.2g\n", nodeList[nodeIndex],
			 currentWave_P->V_fixture_max );
	strcat( inputBuffer, dummyBuffer );
	}
else
	{
	strcpy( inputBuffer, inputBuffer2 );
	sprintf( dummyBuffer, "VFIXS2I %s 0 %6.2g\n", nodeList[nodeIndex],
			 currentWave_P->V_fixture );
	strcat( inputBuffer, dummyBuffer );
	}

if ( setupSpiceInputFile( iterate, spiceType, spiceInFile, headerLine, 
						  spiceFile, model_P->modelFileMax, inputBuffer, 
						  powerBuffer, temperatureBuffer, analysisBuffer ) )
	return( 1 );

callSpice( iterate, spiceCommand, spiceType, spiceInFile, spiceOutFile,
		   spiceMsgFile );

if ( checkForAbort( spiceType, spiceOutFile, spiceMsgFile ) )
	{
	fprintf( stderr, "s2ibis2: Spice run (MAX) aborted.  " );
	fprintf( stderr, "See file %s for information.\n", spiceMsgFile );
	}
else
	getSpiceWaveData( currentWave_P, model_P->simTime, spiceType, 
					  spiceOutFile, "max" ); 

if ( cleanup )
	{
	if ( remove( spiceInFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceInFile );
	if ( remove( spiceOutFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceOutFile );
	if ( remove( spiceMsgFile ) )
		fprintf( stderr, "s2ibis2: Can't remove file %s.\n", spiceMsgFile );
	}
	
return( 0 );
}
